#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import subprocess

import pytest

from build_harness.commands import ExitState, main
from build_harness.commands.analysis import (
    StaticAnalysisError,
    _apply_analysis,
    _run_flake8,
    _run_mypy,
    _run_pydocstyle,
)
from tests.ci.support.click_runner import click_runner  # noqa: F401

MOCK_PROJECT_PATH = pathlib.Path("project/path")
MOCK_VENV_PATH = pathlib.Path("/some/test_static_analysis/path")


class TestRunAnalyses:
    def test_flake8_dirty_exit_raises(self, mocker):
        mock_exit = mocker.MagicMock()
        mock_exit.returncode = 1
        mocker.patch(
            "build_harness.commands.analysis.run_command",
            return_value=mock_exit,
        )

        with pytest.raises(
            StaticAnalysisError, match=r"^flake8 analysis failed"
        ):
            _run_flake8(MOCK_PROJECT_PATH, MOCK_VENV_PATH)

    def test_mypy_dirty_exit_raises(self, mocker):
        mock_exit = mocker.MagicMock()
        mock_exit.returncode = 1
        mocker.patch(
            "build_harness.commands.analysis.run_command",
            return_value=mock_exit,
        )

        with pytest.raises(StaticAnalysisError, match=r"^mypy analysis failed"):
            _run_mypy(MOCK_PROJECT_PATH, MOCK_VENV_PATH)


class TestApplyAnalysis:
    def test_all_called(self, mocker):
        mock_flake8_run = mocker.patch(
            "build_harness.commands.analysis._run_flake8"
        )
        mock_mypy_run = mocker.patch(
            "build_harness.commands.analysis._run_mypy"
        )
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="this/source_dir",
        )

        _apply_analysis("all", MOCK_PROJECT_PATH, MOCK_VENV_PATH)

        mock_flake8_run.assert_called_once_with(
            MOCK_PROJECT_PATH / "this/source_dir", MOCK_VENV_PATH
        )
        mock_mypy_run.assert_called_once_with(
            MOCK_PROJECT_PATH / "this/source_dir", MOCK_VENV_PATH
        )

    def test_flake8_called(self, mocker):
        mock_flake8_run = mocker.patch(
            "build_harness.commands.analysis._run_flake8"
        )
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="this/source_dir",
        )

        _apply_analysis("flake8", MOCK_PROJECT_PATH, MOCK_VENV_PATH)

        mock_flake8_run.assert_called_once_with(
            MOCK_PROJECT_PATH / "this/source_dir", MOCK_VENV_PATH
        )

    def test_mypy_called(self, mocker):
        mock_mypy_run = mocker.patch(
            "build_harness.commands.analysis._run_mypy"
        )
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="this/source_dir",
        )

        _apply_analysis("mypy", MOCK_PROJECT_PATH, MOCK_VENV_PATH)

        mock_mypy_run.assert_called_once_with(
            MOCK_PROJECT_PATH / "this/source_dir", MOCK_VENV_PATH
        )

    def test_pydocstyle_called(self, mocker):
        mock_pydocstyle_run = mocker.patch(
            "build_harness.commands.analysis._run_pydocstyle"
        )
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="this/source_dir",
        )

        _apply_analysis("pydocstyle", MOCK_PROJECT_PATH, MOCK_VENV_PATH)

        mock_pydocstyle_run.assert_called_once_with(
            MOCK_PROJECT_PATH / "this/source_dir", MOCK_VENV_PATH
        )


class TestStaticAnalysis:
    def test_unknown_error(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands.analysis._apply_analysis",
            side_effect=RuntimeError(),
        )
        result = click_runner.invoke(
            main, ["static-analysis", "--analysis", "flake8"]
        )

        assert result.exit_code == ExitState.UNKNOWN_ERROR.value


class TestAll:
    def test_default(self, click_runner, mocker):
        mock_analysis = mocker.patch(
            "build_harness.commands.analysis._apply_analysis"
        )
        mock_formatting = mocker.patch(
            "build_harness.commands.analysis._apply_formatting"
        )

        result = click_runner.invoke(main, ["static-analysis"])

        assert result.exit_code == 0
        mock_formatting.assert_called_once()
        mock_analysis.assert_called_once_with("all", mocker.ANY, mocker.ANY)

    def test_explicit(self, click_runner, mocker):
        mock_analysis = mocker.patch(
            "build_harness.commands.analysis._apply_analysis"
        )
        mock_formatting = mocker.patch(
            "build_harness.commands.analysis._apply_formatting"
        )

        result = click_runner.invoke(
            main, ["static-analysis", "--analysis", "all"]
        )

        assert result.exit_code == 0
        mock_formatting.assert_called_once()
        mock_analysis.assert_called_once_with("all", mocker.ANY, mocker.ANY)

    def test_flake8_failure(self, click_runner, mocker):
        mocker.patch("build_harness.commands.analysis._apply_formatting")
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="something",
        )
        mock_results = [mocker.create_autospec(subprocess.CompletedProcess)]
        mock_results[0].returncode = 1
        mocker.patch(
            "build_harness.commands.analysis.run_command",
            side_effect=mock_results,
        )

        result = click_runner.invoke(main, ["static-analysis"])

        assert result.exit_code == ExitState.FLAKE8_FAILED.value
        assert "flake8 analysis failed" in result.output

    def test_mypy_failure(self, click_runner, mocker):
        mocker.patch("build_harness.commands.analysis._apply_formatting")
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="something",
        )
        mock_results = [
            mocker.create_autospec(subprocess.CompletedProcess),
            mocker.create_autospec(subprocess.CompletedProcess),
        ]
        mock_results[0].returncode = 0
        mock_results[1].returncode = 1
        mocker.patch(
            "build_harness.commands.analysis.run_command",
            side_effect=mock_results,
        )

        result = click_runner.invoke(main, ["static-analysis"])

        assert result.exit_code == ExitState.MYPY_FAILED.value
        assert "mypy analysis failed" in result.output


class TestFlake8:
    def test_clean(self, click_runner, mocker):
        mock_analysis = mocker.patch(
            "build_harness.commands.analysis._apply_analysis"
        )
        mock_formatting = mocker.patch(
            "build_harness.commands.analysis._apply_formatting"
        )

        result = click_runner.invoke(
            main, ["static-analysis", "--analysis", "flake8"]
        )

        assert result.exit_code == 0
        mock_formatting.assert_called_once()
        mock_analysis.assert_called_once_with("flake8", mocker.ANY, mocker.ANY)

    def test_failure(self, click_runner, mocker):
        mocker.patch("build_harness.commands.analysis._apply_formatting")
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="something",
        )
        mock_results = [mocker.create_autospec(subprocess.CompletedProcess)]
        mock_results[0].returncode = 1
        mocker.patch(
            "build_harness.commands.analysis.run_command",
            side_effect=mock_results,
        )

        result = click_runner.invoke(
            main, ["static-analysis", "--analysis", "flake8"]
        )

        assert result.exit_code == ExitState.FLAKE8_FAILED.value
        assert "flake8 analysis failed" in result.output


class TestMypy:
    def test_clean(self, click_runner, mocker):
        mock_analysis = mocker.patch(
            "build_harness.commands.analysis._apply_analysis"
        )
        mock_formatting = mocker.patch(
            "build_harness.commands.analysis._apply_formatting"
        )

        result = click_runner.invoke(
            main, ["static-analysis", "--analysis", "mypy"]
        )

        assert result.exit_code == 0
        mock_formatting.assert_called_once()
        mock_analysis.assert_called_once_with("mypy", mocker.ANY, mocker.ANY)

    def test_failure(self, click_runner, mocker):
        mocker.patch("build_harness.commands.analysis._apply_formatting")
        mocker.patch(
            "build_harness.commands.analysis.acquire_source_dir",
            return_value="something",
        )
        mock_results = [mocker.create_autospec(subprocess.CompletedProcess)]
        mock_results[0].returncode = 1
        mocker.patch(
            "build_harness.commands.analysis.run_command",
            side_effect=mock_results,
        )

        result = click_runner.invoke(
            main, ["static-analysis", "--analysis", "mypy"]
        )

        assert result.exit_code == ExitState.MYPY_FAILED.value
        assert "mypy analysis failed" in result.output



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Manage git command options."""

import dataclasses
import logging
import os
import pathlib
import re
import shutil
import subprocess  # noqa: S404
import typing

import git  # type: ignore
import git.objects  # type: ignore

log = logging.getLogger(__name__)

DEFAULT_DEFAULT_BRANCH_NAME = "master"


class GitError(Exception):
    """Problem with using Git."""


class GitNotFoundError(GitError):
    """Git executable not found error."""


class GitRepoError(GitError):
    """Git repository error."""


class TagDataError(Exception):
    """Tag, offset extraction error."""


def _validate_git_version() -> None:
    result = subprocess.run(  # noqa: S607 S603
        ["git", "--version"], capture_output=True, text=True
    )

    if result.returncode != 0:
        log.error(f"git stdout, {result.stdout}")
        log.error(f"git stderr, {result.stderr}")
        raise GitError("Unable to acquire git version")
    else:
        regex_match = re.match(
            r"git\s+version\s+(?P<major>\d+)\.(?P<minor>\d+).*", result.stdout
        )
        if (
            regex_match
            and (int(regex_match.group("major")) >= 2)
            and (int(regex_match.group("minor")) >= 28)
        ):
            log.info(f"{result.stdout}")
        else:
            raise GitError(
                f"Incorrect git version must be >=2.28, {result.stdout}"
            )


def validate_git(optional_git: typing.Optional[str]) -> None:
    """
    Validate that git is installed or the user has provided a valid git path.

    Args:
        optional_git: Path to git executable.

    Raises:
        GitNotFoundError: If git executable not found.
    """
    system_git = shutil.which("git")
    if optional_git:
        system_git = optional_git

    if not system_git:
        raise GitNotFoundError("Git not available on users PATH")
    if not os.path.exists(system_git):
        raise GitNotFoundError(
            "User specified git invalid, {0}".format(system_git)
        )

    _validate_git_version()


@dataclasses.dataclass
class TagData:
    """Tag, offset data extracted from git describe."""

    tag: str
    offset: typing.Optional[str] = None


def _is_commit_tagged(commit_sha: git.objects.Commit, repo: git.Repo) -> bool:
    """Verify that the specified commit is not tagged."""
    tag_commits = [x.commit for x in repo.tags]
    return any(commit_sha == x for x in tag_commits)


def _parse_describe(describe_output: str) -> TagData:
    """Extract tag, offset data from git describe output."""
    match = re.search(r"(?P<tag>.+?)(-(?P<offset>\d+)-(.+))?$", describe_output)

    if match:
        value = TagData(tag=match.group("tag"), offset=match.group("offset"))
    else:
        raise TagDataError(
            "Unable to match git describe output, {0}".format(describe_output)
        )

    return value


def get_tag_data(
    repo_path: pathlib.Path,
    default_branch_name: typing.Optional[str] = None,
) -> TagData:
    """
    Acquire tag, offset data for release id construction.

    Assumes that the repo has already been set to the required HEAD state.

    Args:
        repo_path: Path to git repository.
        default_branch_name: Name of the default branch in the repo.

    Returns:
        Parsed tag data, if possible.
    Raises:
        GitRepoError: If unable to parse PEP-440 compliant tag.
    """
    try:
        this_repo = git.Repo(str(repo_path))
        try:
            # Start with the nearest historical tag.
            result = this_repo.git.describe("--first-parent", "--tags")

            # If HEAD is dryrun tagged then this is a dryrun so use that tag
            # (ie. treat it as a release).
            if (not _is_commit_tagged(this_repo.head.commit, this_repo)) and (
                re.search(r"\+dryrun", result) is not None
            ):
                # It's a dryrun tag, so search again excluding dryrun tags.
                result = this_repo.git.describe(
                    "--first-parent",
                    "--tags",
                    "--exclude",
                    "*+dryrun*",
                )
            tag_data = _parse_describe(result)
        except git.GitCommandError as e:
            log.debug("handling GitCommandError, {0}".format(str(e)))
            # NOTE: the fallback here is to assume the active branch is the
            #       default branch. there might be unexpected side effects of
            #       this choice.
            this_branch = (
                default_branch_name
                if default_branch_name
                else this_repo.active_branch.name
            )
            number_commits = len(list(this_repo.iter_commits(this_branch)))
            tag_data = TagData(tag="0.0.0", offset=str(number_commits))

    except TagDataError as e:
        log.debug("handling TagDataError, {0}".format(str(e)))
        # extreme fallback that is not expected to happen in practice
        # use a PEP-440 local identifier to prevent publishing because at this
        # point that is almost certainly not desirable
        tag_data = TagData(tag="0.0.0", offset=None)
    except git.InvalidGitRepositoryError as e:
        log.debug("handling git.InvalidGitRepositoryError, {0}".format(str(e)))
        raise GitRepoError(
            "Invalid git repository, {0}".format(repo_path)
        ) from e

    return tag_data



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
import logging
import subprocess

import pytest
from click.testing import CliRunner

import build_harness._logging
import build_harness.commands.build_harness_group
from build_harness.commands import main
from tests.ci.support.click_runner import click_runner  # noqa: F401


@pytest.fixture()
def mock_code_style(mocker):
    mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_isort_result.returncode = 0
    mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_black_result.returncode = 0
    mocker.patch(
        "build_harness.commands.code_style.run_command",
        side_effect=[mock_isort_result, mock_black_result],
    )
    build_harness.commands.build_harness_group.sys.argv = [
        "/some/test_logging_options/path/build-harness"
    ]


@pytest.fixture()
def mock_command_state(mocker):
    mocker.patch("build_harness._logging.logging.handlers.RotatingFileHandler")
    mocker.patch.object(build_harness._logging.logging, "getLogger")
    this_command_state = mocker.patch(
        "build_harness.commands.build_harness_group.CommandState"
    )

    return this_command_state


def test_default_logging(click_runner, mock_code_style, mock_command_state):
    result = click_runner.invoke(main, ["formatting"])
    assert result.exit_code == 0

    mock_command_state.assert_called_once()
    # assume only kwargs are used in the call
    call_args = mock_command_state.call_args_list[0].kwargs
    assert not call_args["logging_state"].configuration.disable_file_logging
    assert not call_args["logging_state"].configuration.enable_console_logging
    assert call_args["logging_state"].configuration.log_level == logging.WARNING


def test_enable_console_logging(
    click_runner, mock_code_style, mock_command_state
):
    result = click_runner.invoke(main, ["--log-enable-console", "formatting"])
    assert result.exit_code == 0

    mock_command_state.assert_called_once()
    # assume only kwargs are used in the call
    call_args = mock_command_state.call_args_list[0].kwargs
    assert not call_args["logging_state"].configuration.disable_file_logging
    assert call_args["logging_state"].configuration.enable_console_logging
    assert call_args["logging_state"].configuration.log_level == logging.WARNING


def test_disable_file_logging(
    click_runner, mock_code_style, mock_command_state
):
    result = click_runner.invoke(main, ["--log-disable-file", "formatting"])
    assert result.exit_code == 0

    mock_command_state.assert_called_once()
    # assume only kwargs are used in the call
    call_args = mock_command_state.call_args_list[0].kwargs
    assert call_args["logging_state"].configuration.disable_file_logging
    assert not call_args["logging_state"].configuration.enable_console_logging
    assert call_args["logging_state"].configuration.log_level == logging.WARNING


class TestLogLevel:
    def _do_test(self, this_level: str, mock_state):
        click_runner = CliRunner()
        result = click_runner.invoke(
            main, ["--log-level", this_level, "formatting"]
        )
        if result.exit_code != 0:
            pytest.fail("Valid log level failed test, {0}".format(this_level))

        mock_state.assert_called_once()
        # assume only kwargs are used in the call
        call_args = mock_state.call_args_list[0].kwargs
        assert call_args["logging_state"].configuration.log_level == getattr(
            logging, this_level.upper()
        )

    def test_critical(self, mock_code_style, mock_command_state):
        self._do_test("critical", mock_command_state)

    def test_error(self, mock_code_style, mock_command_state):
        self._do_test("error", mock_command_state)

    def test_warning(self, mock_code_style, mock_command_state):
        self._do_test("warning", mock_command_state)

    def test_info(self, mock_code_style, mock_command_state):
        self._do_test("info", mock_command_state)

    def test_debug(self, mock_code_style, mock_command_state):
        self._do_test("debug", mock_command_state)

    def test_notset(self, mock_code_style, mock_command_state):
        self._do_test("notset", mock_command_state)

    def test_bad_level(self, click_runner, mock_code_style, mock_command_state):
        result = click_runner.invoke(
            main, ["--log-level", "bad_level", "formatting"]
        )
        assert result.exit_code != 0

    def test_mixed_case(
        self, click_runner, mock_code_style, mock_command_state
    ):
        result = click_runner.invoke(
            main, ["--log-level", "Debug", "formatting"]
        )
        assert result.exit_code == 0

        mock_command_state.assert_called_once()
        # assume only kwargs are used in the call
        call_args = mock_command_state.call_args_list[0].kwargs
        assert (
            call_args["logging_state"].configuration.log_level == logging.DEBUG
        )

    def test_upper_case(
        self, click_runner, mock_code_style, mock_command_state
    ):
        result = click_runner.invoke(
            main, ["--log-level", "DEBUG", "formatting"]
        )
        assert result.exit_code == 0

        mock_command_state.assert_called_once()
        # assume only kwargs are used in the call
        call_args = mock_command_state.call_args_list[0].kwargs
        assert (
            call_args["logging_state"].configuration.log_level == logging.DEBUG
        )



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

from build_harness.tools import Pep440Patterns


class TestPep440:
    def test_regex(self):
        under_test = Pep440Patterns.SPECIFIER_PATTERN

        result = under_test.search("empty_specifier")
        assert result is not None

        result = under_test.search("multiple_specifiers >=2.2, <3.0")
        assert result is not None
        assert result.group("specifier_sets") == ">=2.2, <3.0"

        result = under_test.search("missing_whitespace>=2.2,<3.0")
        assert result is not None
        assert result.group("specifier_sets") == ">=2.2,<3.0"



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pytest

import build_harness.commands.build_harness_group


@pytest.fixture
def mock_sysargv():
    build_harness.commands.build_harness_group.sys.argv = [
        "/some/conftest/path/build-harness"
    ]



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Dependency installation subcommand implementation."""

import io
import logging
import pathlib
import re
import sys
import typing

import click

from build_harness._utility import report_console_error, run_command
from build_harness.project import (
    DEPENDENCY_TYPES,
    DependencyValidationError,
    ProjectDependencies,
    ProjectDependencyAnalysis,
    PyprojecttomlError,
    acquire_project_dependencies,
)
from build_harness.tools import Pep426Patterns, Pep440Patterns, pep503_normalize

from .state import CommandState, ExitState

log = logging.getLogger(__name__)

VALID_DEPENDENCY_CHOICES = [
    "all",
    "dev",
    "runtime",
    "doc",
    "test",
]


class BootstrapError(Exception):
    """Problem occurred during dependency installation."""


class DependencyCheckError(BootstrapError):
    """Installed project dependencies failed to align with declarations."""


class DependencyInstallError(BootstrapError):
    """Problem occurred during pip installation of project dependencies."""


class VenvError(BootstrapError):
    """Problem occurred creating or validating the virtual environment."""


class VenvPipError(BootstrapError):
    """Problem occurred using virtual environment pip."""


PIP_LIST_PATTERN = re.compile(
    r"^\s*{0}\s+(?P<release_id>{1})".format(
        Pep426Patterns.UNNORMALIZED_PACKAGE_NAME,
        Pep440Patterns.VALID_VERSION_CHARACTERS,
    )
)

PipList = typing.List[typing.Tuple[str, str]]


def _extract_pip_list_packages(pip_output: str) -> PipList:
    with io.StringIO(initial_value=pip_output) as f:
        lines = f.readlines()

    pip_list: PipList = []
    for this_line in lines:
        result = PIP_LIST_PATTERN.search(this_line)
        if (
            (result is not None)
            and (result.group("package_name") != "Package")
            and (not result.group("package_name").startswith("--"))
        ):
            pip_list.append(
                (
                    pep503_normalize(result.group("package_name")),
                    result.group("release_id"),
                )
            )

    return pip_list


def _check_pip_versions(
    pip_list: typing.List[
        typing.Tuple[str, str],
    ],
    analysis: ProjectDependencyAnalysis,
) -> None:
    """
    Check that pip installed versions align with project declared dependencies.

    Args:
        pip_list:
        analysis:

    Raises:
        DependencyCheckError: If there is a version mismatch.
    """
    validated_releases = []
    failed_packages = []
    for package_name, package_release in pip_list:
        try:
            validated = analysis.valid_release(package_name, package_release)

            validated_releases.append(validated)
            if not validated:
                failed_packages.append((package_name, package_release))
        except DependencyValidationError:
            # ignore packages that are installed in pip venv, but not declared
            # in project dependencies.
            pass

    if not all(validated_releases):
        raise DependencyCheckError(
            "Installed packages do not comply with declared project "
            "dependencies, {0}".format(failed_packages)
        )


def _check_dependency_names(
    pip_list: typing.List[
        typing.Tuple[str, str],
    ],
    analysis: ProjectDependencyAnalysis,
) -> None:
    """
    Check that project dependencies are installed in the virtual environment.

    Args:
        pip_list: List of packages installed in virtual environment.
        analysis: Project dependencies analysis object.

    Raises:
        DependencyCheckError: If there is a version mismatch.
    """
    project_packages = analysis.packages
    pip_package_names = [pep503_normalize(x) for x, y in pip_list]

    missing_packages = [
        x
        for x in project_packages
        if pep503_normalize(x) not in pip_package_names
    ]
    if len(missing_packages) != 0:
        raise DependencyCheckError(
            "Project dependencies not installed, {0}".format(missing_packages)
        )


def _check_dependencies(
    venv_path: pathlib.Path,
    project_dependencies: ProjectDependencies,
    dependency_type: DEPENDENCY_TYPES,
) -> None:
    """
    Check project dependencies.

    Checks that virtual environment dependencies are aligned with project
    dependencies declarations.

    Args:
        venv_path:
        project_dependencies:
        dependency_type:

    Raises:
        DependencyCheckError: If dependency check fails.
    """
    analysis = ProjectDependencyAnalysis(project_dependencies, dependency_type)

    venv_pip = venv_path / "pip"
    result = run_command(
        [str(venv_pip), "list"],
        capture_output=True,
        text=True,
        universal_newlines=True,
    )
    if result.returncode != 0:
        raise VenvPipError(
            "Failed to list installed virtual environment packages, "
            "{0}".format(venv_path)
        )
    else:
        pip_list = _extract_pip_list_packages(result.stdout)
        _check_pip_versions(pip_list, analysis)
        _check_dependency_names(pip_list, analysis)


def _check_virtual_environment(project_path: pathlib.Path) -> pathlib.Path:
    """
    Check if the necessary virtual environment exists, creating it if needed.

    Args:
        project_path: Project repo root directory.

    Returns:
        Path to virtual environment bin directory.
    """
    expected_venv_path = project_path / ".venv" / "bin"

    if not expected_venv_path.exists():
        result = run_command(
            ["python3", "-m", "venv", str(expected_venv_path.parent)],
            capture_output=True,
            universal_newlines=True,
        )

        if result.returncode != 0:
            message = "Failed creating virtual environment, {0}".format(
                expected_venv_path.parent
            )
            log.error(message)
            if result.stdout:
                log.error("Venv stdout, {0}".format(result.stdout))
            if result.stderr:
                log.error("Venv stderr, {0}".format(result.stderr))
            raise VenvError(message)
    elif not (expected_venv_path / "pip").is_file():
        # Assume it's an invalid virtual environment
        message = "Invalid virtual environment, {0}".format(expected_venv_path)
        raise VenvError(message)

    return expected_venv_path


def _install_project_dependencies(
    dependencies: typing.List[str], venv_path: pathlib.Path
) -> None:
    """
    Install the specified project dependencies to the virtual environment.

    Args:
        dependencies: List of project dependencies to be installed.
        venv_path: Path to virtual environment bin directory.

    Raises:
        DependencyInstallError: If installation does not exit cleanly.
    """
    dependency_input = "\n".join(dependencies)
    result = run_command(
        [str(venv_path / "pip"), "install", "-r", "/dev/stdin"],
        capture_output=True,
        input=dependency_input,
        text=True,
        universal_newlines=True,
    )

    if result.returncode != 0:
        message = "Dependency installation failed"
        log.error(message)
        log.error("Pip install stdout, {0}".format(result.stdout))
        log.error("Pip install stderr, {0}".format(result.stderr))
        raise DependencyInstallError(message)


def _install_all_project_dependencies(
    project_dependencies: ProjectDependencies, venv_path: pathlib.Path
) -> None:
    """
    Install *all* the project dependencies to the virtual environment.

    Args:
        project_dependencies: Dictionary of project dependencies.
        venv_path: Path to virtual environment bin directory.

    Raises:
        DependencyInstallError: If installation does not exit cleanly.
    """
    aggregate: typing.List[str] = []
    for this_values in project_dependencies.values():
        aggregate += this_values  # type: ignore

    _install_project_dependencies(aggregate, venv_path)


@click.command()
@click.pass_context
@click.option(
    "--check",
    "check_dependencies",
    default=False,
    help="Don't install dependencies, just check if dependencies installed are "
    "consistent with the project requirements (pyproject.toml or "
    "requirements.txt).",
    is_flag=True,
)
@click.option(
    "--dependencies",
    "dependency_type",
    default="all",
    help="Install project package dependencies into virtual environment, "
    "creating the virtual environment if necessary.",
    show_default=True,
    type=click.Choice(VALID_DEPENDENCY_CHOICES, case_sensitive=True),
)
def install(
    ctx: click.Context,
    check_dependencies: bool,
    dependency_type: DEPENDENCY_TYPES,
) -> None:
    """
    Install project dependencies.

    If  virtual environment does not presently exist, one will be created.
    """
    try:
        command_state: CommandState = ctx.obj

        project_dependencies = acquire_project_dependencies(
            command_state.project_path
        )

        # Cannot assume venv_path of this build_harness call is in or related
        # to the target project; user might be bootstrapping using build_
        # harness from another virtual environment.
        venv_path = _check_virtual_environment(command_state.project_path)

        if not check_dependencies:
            if dependency_type == "all":
                _install_all_project_dependencies(
                    project_dependencies, venv_path
                )
            else:
                _install_project_dependencies(
                    # mypy and dynamic data on TypedDict hints again
                    project_dependencies[dependency_type],  # type: ignore
                    venv_path,
                )
        else:
            _check_dependencies(
                venv_path, project_dependencies, dependency_type
            )
    except PyprojecttomlError as e:
        message = str(e)
        report_console_error(message)
        sys.exit(ExitState.BAD_PYPROJECTTOML.value)
    except (DependencyInstallError, VenvError) as e:
        message = str(e)
        report_console_error(message)
        sys.exit(ExitState.DEPENDENCY_INSTALL_FAILED.value)
    except (DependencyCheckError, VenvPipError) as e:
        message = str(e)
        report_console_error(message)
        sys.exit(ExitState.DEPENDENCY_CHECK_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pytest
from cookiecutter.main import cookiecutter

from tests.ci.support.project_context import (
    build_cookiecutter_context,
    make_project_venv,
    run_command,
)


class TestDefaultBranchCookiecutter:
    def test_default_branches(self):
        branches = ["main", "master"]

        with build_cookiecutter_context() as this_context:
            for this_branch in branches:
                project_name = this_branch
                context_configuration = {
                    "build_harness_version": "==0.0.0",
                    "default_branch": this_branch,
                    "project_name": project_name,
                    "disable_bh_dependencies": True,
                }

                project_dir = make_project_venv(
                    project_name, this_context.temp_path
                )

                cookiecutter(
                    str(this_context.template_dir),
                    extra_context=context_configuration,
                    no_input=True,
                    replay=False,
                    overwrite_if_exists=True,
                    output_dir=this_context.temp_path,
                )

                response = run_command(
                    ["git", "-C", str(project_dir), "branch", "-l"],
                    capture_output=True,
                    text=True,
                )
                if response.returncode != 0:
                    pytest.fail("git command failed, {0}".format(this_branch))

                if "* {0}".format(this_branch) not in response.stdout:
                    pytest.fail(
                        "{0} branch not in repo, {1}".format(
                            this_branch, response.stdout
                        )
                    )



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import os
import pathlib

from behave import then, when

from build_harness._utility import run_command

from .support import build_harness_context, chdir_project_dir, create_venv

DEBUG_INSTALL = False


@when("the build harness command is run")
def step_impl(context):
    pre_step_actions = None
    if hasattr(context, "pre_step_actions"):
        pre_step_actions = context.pre_step_actions
    release_id = "0.0.0"
    if hasattr(context, "release_id"):
        release_id = context.release_id

    test_install = False
    if hasattr(context, "test_install"):
        test_install = True

    with build_harness_context(
        context.create_pyproject,
        release_id,
        context.existing_project_venv,
        debug_install=DEBUG_INSTALL,
        pre_step_actions=pre_step_actions,
    ) as (
        temporary_dir,
        this_bh_command,
    ):
        mock_project = pathlib.Path(temporary_dir)
        print(f"project dir, {mock_project}")
        print(f"cwd before chdir, {os.getcwd()}")
        with chdir_project_dir(mock_project):
            print(f"cwd after chdir, {os.getcwd()}")

            this_command = [str(this_bh_command)] + context.given_arguments
            if hasattr(context, "release_id"):
                this_command += ["--release-id", release_id]
            result = run_command(
                this_command,
                capture_output=True,
                text=True,
                universal_newlines=True,
            )

            print(result.stdout)
            print(result.stderr)

            context.run_result = result

            context.mock_project = mock_project
            context.project_dir_list = list(mock_project.iterdir())
            dist_dir = mock_project / "dist"
            context.dist_dir_list = (
                list(dist_dir.iterdir()) if dist_dir.is_dir() else list()
            )

            if hasattr(context, "post_step_actions"):
                context.post_step_actions(context, result, mock_project)

            if test_install:
                tmpv_pip = create_venv(mock_project / "tmpv")
                this_command = [
                    str(tmpv_pip),
                    "install",
                    str(context.dist_dir_list[1]),
                ]
                result = run_command(
                    this_command,
                    capture_output=True,
                    text=True,
                    universal_newlines=True,
                )

                print(result.stdout)
                print(result.stderr)

                result = run_command(
                    [tmpv_pip, "list"],
                    capture_output=True,
                    text=True,
                    universal_newlines=True,
                )

                context.pip_list = result.stdout


@then("the utility exits clean")
def step_impl(context):
    print(context.run_result.returncode)
    assert context.run_result.returncode == 0


@then("the utility exits dirty")
def step_impl(context):
    print(context.run_result.returncode)
    assert context.run_result.returncode != 0



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Define flit entry points."""

from .commands import main, publish_flow_main, release_flow_main


def build_harness_entry() -> None:
    """Flit entry point to ``build-harness`` command CLI argument parsing."""
    main()


def publish_flow_entry() -> None:
    """Flit entry point to ``publish-flow`` command CLI argument parsing."""
    publish_flow_main()


def release_flow_entry() -> None:
    """Flit entry point to ``release-flow`` command CLI argument parsing."""
    release_flow_main()



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import logging
import pathlib
import typing

import pytest

from build_harness._logging import LoggingState
from build_harness.commands._declarations import (
    DEFAULT_LOG_FILE,
    VALID_LOG_LEVELS,
)
from tests.ci.support.captured_io import capture_io
from tests.ci.support.directory import change_directory


class TestLoggingState:
    def test_disable_console_logging(self):
        arguments = {
            "disable_file_logging": True,
            "enable_console_logging": False,
            "log_level_text": "warning",
        }
        captured_err, captured_out, _ = self._do_test(arguments)

        assert not captured_err
        assert not captured_out

    def _do_test(
        self, arguments: dict
    ) -> typing.Tuple[str, str, typing.Optional[str]]:
        with capture_io() as captured, change_directory() as temp_dir:
            expected_log_file: pathlib.Path = temp_dir / "build_harness.log"
            under_test = LoggingState(**arguments)

            log = logging.getLogger("MockLogging")
            log.debug("debug message")
            log.info("info message")
            log.warning("warning message")
            log.error("error message")
            log.critical("critical message")

            assert (
                under_test.configuration.disable_file_logging
                == arguments["disable_file_logging"]
            )
            assert (
                under_test.configuration.enable_console_logging
                == arguments["enable_console_logging"]
            )
            assert under_test.configuration.log_level == getattr(
                logging, arguments["log_level_text"].upper()
            )

            if arguments["disable_file_logging"]:
                assert under_test.configuration.log_file_path is None
                assert not expected_log_file.exists()

                file_contents = None
            else:
                assert under_test.configuration.log_file_path == pathlib.Path(
                    DEFAULT_LOG_FILE
                )
                assert expected_log_file.is_file()

                with expected_log_file.open("r") as f:
                    file_contents = f.read()

            captured.err.seek(0)
            captured.out.seek(0)

            captured_err = captured.err.read()
            captured_out = captured.out.read()

        return captured_err, captured_out, file_contents

    def test_enable_console_logging(self):
        arguments = {
            "disable_file_logging": True,
            "enable_console_logging": True,
            "log_level_text": "warning",
        }
        captured_err, captured_out, _ = self._do_test(arguments)

        assert not captured_out
        assert "debug" not in captured_err
        assert "info" not in captured_err
        assert "warning message" in captured_err
        assert "error message" in captured_err
        assert "critical message" in captured_err

    def test_disable_file_logging(self):
        arguments = {
            "disable_file_logging": True,
            "enable_console_logging": False,
            "log_level_text": "warning",
        }
        captured_err, captured_out, _ = self._do_test(arguments)

        assert not captured_err
        assert not captured_out

    def test_enable_file_logging(self):
        arguments = {
            "disable_file_logging": False,
            "enable_console_logging": False,
            "log_level_text": "warning",
        }
        captured_err, captured_out, file_contents = self._do_test(arguments)

        assert not captured_err
        assert not captured_out

        assert "debug" not in file_contents
        assert "info" not in file_contents
        assert "warning message" in file_contents
        assert "error message" in file_contents
        assert "critical message" in file_contents

    def test_enable_both_file_console(self):
        arguments = {
            "disable_file_logging": False,
            "enable_console_logging": True,
            "log_level_text": "warning",
        }
        captured_err, captured_out, file_contents = self._do_test(arguments)

        assert "debug" not in captured_err
        assert "info" not in captured_err
        assert "warning message" in captured_err
        assert "error message" in captured_err
        assert "critical message" in captured_err

        assert "debug" not in file_contents
        assert "info" not in file_contents
        assert "warning message" in file_contents
        assert "error message" in file_contents
        assert "critical message" in file_contents

    def test_log_levels(self):
        for this_level in VALID_LOG_LEVELS:
            arguments = {
                "disable_file_logging": True,
                "enable_console_logging": True,
                "log_level_text": this_level,
            }
            captured_err, captured_out, _ = self._do_test(arguments)

            if captured_out:
                pytest.fail("Captured stdout should be empty")

            if this_level != "notset":
                expected_level = this_level.upper()
                if (expected_level not in captured_err) and (
                    "{0} message".format(expected_level.lower())
                    not in captured_err
                ):
                    pytest.fail(
                        "Log level {0} not present in captured stderr, "
                        "{1}".format(
                            expected_level,
                            captured_err if captured_err else "<empty>",
                        )
                    )
            else:
                # notset enables logging at all levels
                for log_level in VALID_LOG_LEVELS:
                    if log_level != "notset":
                        assert "{0} message".format(log_level) in captured_err



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""General support for build harness implementation."""

import logging
import pathlib
import re
import subprocess  # noqa: S404
import typing

import click

log = logging.getLogger(__name__)

CommandArgs = typing.List[str]


def command_path(venv_path: pathlib.Path, command_array: CommandArgs) -> str:
    """Construct command path, including virtual environment path."""
    this_path = str(venv_path / command_array[0])
    log.debug("command path, {0}".format(this_path))

    return this_path


def run_command(
    command: CommandArgs,
    suppress_command_logging: bool = False,
    suppress_argument_logging: bool = False,
    **kwargs: typing.Any,
) -> subprocess.CompletedProcess:
    """
    Run a system command using ``subprocess.run``.

    Args:
        command: List of command and arguments.
        suppress_command_logging: Suppress logging of command.
        suppress_argument_logging: Suppress logging of subprocess run command
                                   arguments.
        **kwargs: Optional arguments passed through to ``subprocess.run``.

    Returns:
        Subprocess results.
    """
    if not suppress_command_logging:
        log.debug("command to run, {0}".format(str(command)))
    if not suppress_argument_logging:
        log.debug("command arguments, {0}".format(str(kwargs)))
    final_args = {
        **kwargs,
        **{
            # always return text as string, not bytes
            "text": True,
        },
    }
    result = subprocess.run(command, **final_args)  # noqa: S603 S604

    log.debug("command return code, {0}".format(result.returncode))
    log.debug("command stdout, {0}".format(result.stdout))
    log.debug("command stderr, {0}".format(result.stderr))

    return result


def extract_coverage(report: str) -> int:
    """
    Extract total coverage result from a coverage report.

    Args:
        report: IO stream of captured coverage report output.

    Returns:
        Extracted total reported coverage.
    """
    current_coverage = 0
    lines = report.splitlines()
    for this_line in lines:
        match_result = re.search(
            r"^\s*((TOTAL)|((T|t)otal)).*\s+(?P<total>\d{1,3})%\s*$", this_line
        )  # type: ignore
        if match_result is not None:
            log.debug("coverage report line total, {0}".format(this_line))
            current_coverage = int(match_result.group("total"))  # type: ignore

    return current_coverage


def report_console_error(message: str) -> None:
    """
    Report an error to the click console.

    Args:
        message: Message to be reported.
    """
    click.echo(
        click.style("FAILED: {0}".format(message), bold=True, fg="red"),
        err=True,
    )



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Unit tests subcommand implementation."""

import logging
import pathlib
import sys
import typing

import click

from build_harness._utility import (
    extract_coverage,
    report_console_error,
    run_command,
)
from build_harness.project import acquire_source_dir
from build_harness.tools import PytestCommand, TestCommandOptions

from .state import CommandState, ExitState

log = logging.getLogger(__name__)

DEFAULT_OUTPUT_DIR = "dist"
DEFAULT_TEST_DIR = "tests"

COVERAGE_REPORT_ARGUMENTS = [
    "--coverage-console",
    "--coverage-html",
    "--coverage-xml",
]


class UnitTestError(Exception):
    """Problem occurred during running unit tests."""


class TestFailureError(UnitTestError):
    """At least one unit test failed."""


class TestCoverageError(UnitTestError):
    """Test coverage failed to meet or exceed the defined threshold."""


def _run_tests(
    command_options: TestCommandOptions,
    coverage_threshold: typing.Optional[int],
    coverage_console: bool,
) -> None:
    """
    Run the unit tests.

    Args:
        command_options: pytest command options.
        coverage_threshold: If defined, the coverage threshold to test against.
        coverage_console: Enable coverage report to console.
    """
    this_pytest = PytestCommand(command_options)

    result = run_command(
        this_pytest.command(),
        capture_output=True,
    )
    if (result.returncode not in [0, 5]) or (
        (result.returncode == 5) and (not command_options["pass_zero_tests"])
    ):
        message = "unit tests failed. See log for details"
        log.debug(message)
        raise TestFailureError(message)
    elif result.returncode == 5:
        log.warning("Zero tests found, but options require exiting clean")

    if (coverage_threshold is not None) and result.stdout:
        report_text = result.stdout
        log.debug("coverage report, {0}".format(report_text))
        actual_coverage = extract_coverage(report_text)
        if actual_coverage < coverage_threshold:
            message = (
                "Coverage test failed, {0} (actual) < {1} (threshold)".format(
                    actual_coverage, coverage_threshold
                )
            )
            log.debug(message)
            raise TestCoverageError(message)
        else:
            click.echo("Total coverage, {0}".format(actual_coverage))
    elif (coverage_threshold is not None) and (not result.stdout):
        message = "no coverage report captured. See log for details"
        log.debug(message)
        raise TestCoverageError(message)

    if coverage_console and result.stdout:
        click.echo(result.stdout)


def _process_html_coverage(
    coverage_html: bool,
    coverage_html_dir: typing.Optional[str],
    output_path: pathlib.Path,
) -> typing.Tuple[bool, typing.Optional[pathlib.Path]]:
    html_coverage_path: typing.Optional[pathlib.Path]

    html_coverage_enabled = True
    if (not coverage_html_dir) and coverage_html:
        html_coverage_path = output_path / "coverage_report"
    elif coverage_html_dir:
        html_coverage_path = pathlib.Path(coverage_html_dir)
    else:
        html_coverage_enabled = False
        html_coverage_path = None

    log.debug("HTML coverage enabled, {0}".format(str(html_coverage_enabled)))
    log.debug("HTML coverage path, {0}".format(str(html_coverage_path)))

    return html_coverage_enabled, html_coverage_path


def _process_xml_coverage(
    coverage_xml: bool,
    coverage_xml_file: typing.Optional[str],
    output_path: pathlib.Path,
) -> typing.Tuple[bool, typing.Optional[pathlib.Path]]:
    xml_coverage_path: typing.Optional[pathlib.Path]

    xml_coverage_enabled = True
    if (not coverage_xml_file) and coverage_xml:
        xml_coverage_path = output_path / "coverage_report.xml"
    elif coverage_xml_file:
        xml_coverage_path = pathlib.Path(coverage_xml_file)
    else:
        xml_coverage_enabled = False
        xml_coverage_path = None

    log.debug("XML coverage enabled, {0}".format(str(xml_coverage_enabled)))
    log.debug("XML coverage path, {0}".format(str(xml_coverage_path)))

    return xml_coverage_enabled, xml_coverage_path


def _process_junitxml(
    junitxml: bool,
    junitxml_file: typing.Optional[str],
    output_path: pathlib.Path,
) -> typing.Tuple[bool, typing.Optional[pathlib.Path]]:
    junitxml_path: typing.Optional[pathlib.Path]

    junitxml_enabled = True
    if (not junitxml_file) and junitxml:
        junitxml_path = output_path / "junit_report.xml"
    elif junitxml_file:
        junitxml_path = pathlib.Path(junitxml_file)
    else:
        junitxml_enabled = False
        junitxml_path = None

    log.debug(
        "JUnit XML test report enabled, {0}".format(str(junitxml_enabled))
    )
    log.debug("JUnit XML test report path, {0}".format(str(junitxml_path)))

    return junitxml_enabled, junitxml_path


@click.command()
@click.pass_context
@click.option(
    "--check",
    "coverage_check",
    default=None,
    help="Test coverage against the specified threshold. [default: disabled]",
    type=int,
)
@click.option(
    COVERAGE_REPORT_ARGUMENTS[0],
    default=False,
    help="Enable console coverage report.",
    is_flag=True,
    show_default=True,
)
@click.option(
    COVERAGE_REPORT_ARGUMENTS[1],
    default=False,
    help="Enable HTML coverage report.",
    is_flag=True,
)
@click.option(
    "{0}-file".format(COVERAGE_REPORT_ARGUMENTS[1]),
    default="",
    help="Enable HTML coverage report. [default: auto directory location when "
    "enabled]",
    metavar="DIRECTORY",
    type=str,
)
@click.option(
    COVERAGE_REPORT_ARGUMENTS[2],
    default=False,
    help="Enable XML coverage report.",
    is_flag=True,
    type=str,
)
@click.option(
    "{0}-file".format(COVERAGE_REPORT_ARGUMENTS[2]),
    default="",
    help="Enable XML coverage report. [default: auto file location, when "
    "enabled]",
    metavar="FILE",
    show_default=True,
    type=str,
)
@click.option(
    "--junitxml",
    default=False,
    help="Enable junit XML test report.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--junitxml-file",
    default=None,
    help="Path to junit XML test report. [default: disabled, default auto file "
    "location when enabled]",
    metavar="FILE",
    show_default=True,
    type=str,
)
@click.option(
    "--output-dir",
    default=DEFAULT_OUTPUT_DIR,
    help="Path to directory for placing output artifacts such as coverage "
    "report files.",
    metavar="DIRECTORY",
    show_default=True,
    type=str,
)
@click.option(
    "--pass-zero-tests",
    default=False,
    help="Pass (exit clean) when there are zero tests found.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--source-dir",
    default="",
    help="Path to source directory. [default: acquire from project]",
    metavar="DIRECTORY",
    type=str,
)
@click.option(
    "--test-dir",
    default=DEFAULT_TEST_DIR,
    help="Path to unit test directory.",
    metavar="DIRECTORY",
    show_default=True,
    type=str,
)
def unit_test(
    ctx: click.Context,
    coverage_check: typing.Optional[int],
    coverage_console: bool,
    coverage_html: bool,
    coverage_html_file: typing.Optional[str],
    coverage_xml: bool,
    coverage_xml_file: typing.Optional[str],
    junitxml: bool,
    junitxml_file: typing.Optional[str],
    output_dir: str,
    pass_zero_tests: bool,
    source_dir: str,
    test_dir: str,
) -> None:
    """
    Run tests using the ``pytest`` unit test framework.

    When enabled, any reports will be automatically placed in the output
    artifact directory unless an alternative location is specified to the
    relevant argument.
    """
    try:
        command_state: CommandState = ctx.obj

        output_path = pathlib.Path(output_dir)

        if not source_dir:
            source_dir = acquire_source_dir(pathlib.Path.cwd())

        html_coverage_enabled, html_coverage_path = _process_html_coverage(
            coverage_html, coverage_html_file, output_path
        )
        xml_coverage_enabled, xml_coverage_path = _process_xml_coverage(
            coverage_xml, coverage_xml_file, output_path
        )
        junitxml_enabled, junitxml_path = _process_junitxml(
            junitxml, junitxml_file, output_path
        )

        if coverage_check:
            coverage_console = True

        command_options: TestCommandOptions = {
            "output_path": output_path,
            "source_path": pathlib.Path(source_dir),
            "test_path": pathlib.Path(test_dir),
            "venv_path": command_state.venv_path,
            "pass_zero_tests": pass_zero_tests,
            "report_enabled": {
                "html": html_coverage_enabled,
                "term-missing": coverage_console,
                "xml": xml_coverage_enabled,
                "junitxml": junitxml_enabled,
            },
            "report_dirs": {
                "html": html_coverage_path,
                "term-missing": None,
                "xml": xml_coverage_path,
                "junitxml": junitxml_path,
            },
        }

        _run_tests(command_options, coverage_check, coverage_console)
    except TestCoverageError:
        report_console_error("Test coverage has failed.")
        sys.exit(ExitState.TEST_COVERAGE_FAILED.value)
    except TestFailureError:
        report_console_error("Tests have failed.")
        sys.exit(ExitState.TESTS_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import re

from behave import given, then

from .support import create_mock_pyproject, create_source_dir

DEBUG_INSTALL = False


def package_post_step_actions(context, result, mock_project):
    if result.returncode == 0:
        whl_package_dir = mock_project / "dist" / ""
        context.whl_packages = list(whl_package_dir.iterdir())


@then("the wheel package is created")
def step_impl(context):
    expected_package = "some_module-0.0.0-py2.py3-none-any.whl"
    search_result = [expected_package in str(x) for x in context.whl_packages]
    if not any(search_result):
        print(context.whl_packages)
    assert any(search_result)


def _create_bad_pyproject(file_path: pathlib.Path):
    with file_path.open(mode="w") as f:
        f.write(
            """[build-system]
requires = ["something_else >=2,<3"]
build-backend = "flit_core.buildapi"

[tool.flit.metadata]
module = "some_module"
author = "Some Name"
author-email = "name@somewhere.com"
"""
        )


@given("a flit project directory to be packaged")
def step_impl(context):
    context.existing_project_venv = False
    context.create_pyproject = create_mock_pyproject
    context.given_arguments = ["package"]
    context.pre_step_actions = create_source_dir
    context.post_step_actions = package_post_step_actions
    context.packager = "flit"


@given("a project directory using a non-supported packaging utility")
def step_impl(context):
    context.existing_project_venv = False
    context.create_pyproject = create_mock_pyproject
    context.pre_step_actions = create_source_dir
    context.packager = "something_else"
    context.given_arguments = ["package"]
    context.post_step_actions = package_post_step_actions


@then("the correct packaging tool cannot be identified")
def step_impl(context):
    assert "" in context.run_result.output


@given("a PEP440 compliant release id")
def step_impl(context):
    context.release_id = "3.1.4.dev123"
    context.test_install = True


@then("the wheel package file identifies itself with the release id")
def step_impl(context):
    print(context.dist_dir_list)
    assert any(
        [
            re.search(r"dist/some_module[_\-]3\.1\.4\.dev123-py.+whl", str(x))
            is not None
            for x in context.dist_dir_list
        ]
    )


@then("when installed the package identifies itself with the release id")
def step_impl(context):
    print(context.pip_list)
    assert (
        re.search(r"some_module\s+3\.1\.4\.dev123", context.pip_list)
        is not None
    )


@given("no release id specified")
def step_impl(context):
    assert not hasattr(context, "release_id")


@then("the wheel package file identifies itself with the default release id")
def step_impl(context):
    assert any(
        [
            re.search(r"dist/some_module[_\-]0\.0\.0-py.+whl", str(x))
            is not None
            for x in context.dist_dir_list
        ]
    )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import logging
import pathlib
import subprocess
import sys
import typing

log = logging.getLogger(__name__)

CommandArgs = typing.List[str]

def run_command(
    command: CommandArgs, **kwargs: typing.Any
) -> subprocess.CompletedProcess:
    """
    Run a system command using ``subprocess.run``.

    Args:
        command: List of command and arguments.
        **kwargs: Optional arguments passed through to ``subprocess.run``.

    Returns:
        Subprocess results.
    """
    log.debug("command to run, {0}".format(str(command)))
    log.debug("command arguments, {0}".format(str(kwargs)))
    result = subprocess.run(command, **kwargs)

    return result


def run_command_sequence(sequence: typing.List[str]) -> None:
    """
    Run through a shell command sequence.

    Args:
        sequence: List of commands to be executed.
    """
    for this_command in sequence:
        response = run_command(this_command.split())


def create_venv() -> None:
    """Create virtual environment for new project."""
    venv_dir = pathlib.Path("{{ cookiecutter._venv_bin }}").parent
    commands = [
        "python3 -m venv {0}".format(venv_dir),
    ]
    if "{{ ('disable_bh_dependencies' not in cookiecutter) or (not cookiecutter.disable_bh_dependencies) }}":
        # default install build_harness utility
        commands += [
            "{{ cookiecutter._venv_bin }}/pip install build_harness",
        ]
    else:
        # assume build_harness utility has been installed externally
        if not (venv_dir/"bin/build-harness").is_file():
            raise RuntimeError("Build harness must be installed before bootstrap run "
                               "disabling dependencies.")

    run_command_sequence(commands)


def verify_flit_project_summary() -> None:
    project_summary = "{{ cookiecutter.project_summary }}"

    if not project_summary.endswith("."):
        print(
            "ERROR: flit project summary must end with '.', {0}".format(project_summary)
        )
        sys.exit(1)


def verify_flit() -> None:
    verify_flit_project_summary()


def main() -> None:
    packaging_provider = "{{ cookiecutter.packaging_provider }}"

    create_venv()
    if packaging_provider == "flit":
        verify_flit()


if __name__ == "__main__":
    main()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Manage pytest command options."""

import logging
import pathlib
import typing

from build_harness._utility import command_path

log = logging.getLogger(__name__)

VALID_COVERAGE_REPORTS = ["html", "term-missing", "xml"]

ReportFlags = typing.TypedDict(
    "ReportFlags",
    {
        "junitxml": bool,
        "html": bool,
        "term-missing": bool,
        "xml": bool,
    },
)

ReportPaths = typing.TypedDict(
    "ReportPaths",
    {
        "junitxml": typing.Optional[pathlib.Path],
        # Coverage reports
        "html": typing.Optional[pathlib.Path],
        "term-missing": None,
        "xml": typing.Optional[pathlib.Path],
    },
)


class TestCommandOptions(typing.TypedDict):
    """``pytest`` command options dictionary type hint."""

    output_path: pathlib.Path
    source_path: pathlib.Path
    test_path: pathlib.Path
    venv_path: pathlib.Path

    pass_zero_tests: bool

    report_enabled: ReportFlags
    report_dirs: ReportPaths


T = typing.TypeVar("T", bound="PytestCommand")


class PytestCommand:
    """Manage pytest command line options."""

    COVERAGE_REPORTS = [
        "html",
        "term-missing",
        "xml",
    ]

    def __init__(self: T, command_options: TestCommandOptions) -> None:
        """Construct ``PytestCommand`` object."""
        self.options = command_options

    def command(self: T) -> typing.List[str]:
        """
        Construct pytest command list with argument options.

        Returns:
            Constructed list.
        """
        this_command = [
            "pytest",
        ]

        if any(
            [
                self.options["report_enabled"][x]  # type:ignore
                for x in self.options["report_enabled"].keys()
            ]
        ):
            this_command.append(
                "--cov={0}".format(str(self.options["source_path"]))
            )

        for report, enabled in sorted(self.options["report_enabled"].items()):
            dirs_value = self.options["report_dirs"][report]  # type:ignore
            if enabled and (report in self.COVERAGE_REPORTS):
                this_command.append("--cov-report")
                argument = (
                    "{0}".format(report)
                    if not dirs_value
                    else "{0}:{1}".format(report, dirs_value)
                )
                this_command.append(argument)
            elif enabled and (report == "junitxml"):
                argument = (
                    "--junitxml"
                    if not dirs_value
                    else "--junitxml={0}".format(dirs_value)
                )
                this_command.append(argument)

        this_command.append(str(self.options["test_path"]))

        this_command[0] = command_path(self.options["venv_path"], this_command)
        return this_command



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#


"""Acceptance tests subcommand implementation."""

import logging
import pathlib
import sys
import typing

import click

from build_harness._base_exception import BuildHarnessError
from build_harness._utility import report_console_error
from build_harness.commands.state import CommandState, ExitState

from .run_tests import AcceptanceTestsError, _run_bdd
from .snippets import _generate_snippets
from .tags import _summarize_tags

log = logging.getLogger(__name__)

DEFAULT_JUNITXML_DIR = "dist/behave"

VALID_BDD_RUN_TYPES = ["snippets", "tags", "tests"]


class AcceptanceError(BuildHarnessError):
    """Problem occurred during BDD run."""


@click.command()
@click.pass_context
@click.argument(
    "run_type",
    type=click.Choice(VALID_BDD_RUN_TYPES),
)
@click.option(
    "--junitxml",
    default=False,
    help="Enable junit XML test report.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--junitxml-dir",
    default=None,
    help="Path to junit XML test report directory. [default: default auto file "
    "location when enabled]",
    metavar="FILE",
    type=str,
)
@click.option(
    "--tags",
    default=[None],
    help="Specify tags in Python behave style to select for in a run. "
    "[default: disabled]",
    metavar="BEHAVE_TAG",
    multiple=True,
    type=str,
)
def bdd_acceptance_command(
    ctx: click.Context,
    junitxml: bool,
    junitxml_dir: typing.Optional[str],
    run_type: str,
    tags: typing.Tuple[typing.Optional[str]],
) -> None:
    """
    BDD acceptance tests.

      snippets:  create snippets for BDD statement with no step implementation.
      tags:      show a table showing which feature files tags are used in.
      tests:     run acceptance tests.
    """
    try:
        command_state: CommandState = ctx.obj

        if run_type == "snippets":
            _generate_snippets(command_state.venv_path)
        elif run_type == "tags":
            _summarize_tags(command_state.venv_path)
        elif run_type == "tests":
            if junitxml and junitxml_dir:
                junitxml_path = pathlib.Path(junitxml_dir)
            else:
                junitxml_path = pathlib.Path(DEFAULT_JUNITXML_DIR)
            _run_bdd(command_state.venv_path, junitxml, junitxml_path, tags)
        else:
            raise AcceptanceError(
                "Invalid acceptance run type, {0}".format(run_type)
            )
    except AcceptanceError as e:
        message = str(e)
        report_console_error(message)
        sys.exit(ExitState.ACCEPTANCE_RUN_FAILED.value)
    except AcceptanceTestsError as e:
        message = str(e)
        report_console_error(message)
        sys.exit(ExitState.ACCEPTANCE_TESTS_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

from build_harness.tools.git import _validate_git_version


class TestValidateGitVersion:
    def test_clean(self):
        # WARNING: assumes the test run context includes a valid git release
        # (should be true for developers and CI).
        _validate_git_version()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#


"""Acceptance tests subcommand implementation."""

import logging
import pathlib
import typing

from build_harness._base_exception import BuildHarnessError
from build_harness._utility import command_path, run_command

log = logging.getLogger(__name__)


class AcceptanceTestsError(BuildHarnessError):
    """Problem occurred during acceptance tests run."""


def _check_junit_dir(
    junit_xml_enabled: bool, junitxml_dir: pathlib.Path
) -> None:
    if junit_xml_enabled and (not junitxml_dir.exists()):
        junitxml_dir.mkdir(parents=True)
    elif junit_xml_enabled and (not junitxml_dir.is_dir()):
        raise AcceptanceTestsError(
            "JUnit XML location must be a directory, {0}".format(junitxml_dir)
        )


def _run_bdd(
    venv_path: pathlib.Path,
    junit_xml_enabled: bool,
    junitxml_dir: pathlib.Path,
    tags: typing.Tuple[typing.Optional[str]],
) -> None:
    _check_junit_dir(junit_xml_enabled, junitxml_dir)

    behave_command = [
        "behave",
        "--format",
        "pretty",
    ]

    if junit_xml_enabled:
        behave_command += [
            "--junit",
            "--junit-directory",
            str(junitxml_dir),
        ]

    behave_command += [
        "--tags",
        "~@notimplemented",
        "--tags",
        "~@nottestable",
    ]

    for this_tags in tags:
        if this_tags:
            behave_command += [
                "--tags",
                this_tags,
            ]

    behave_command += [
        "features",
    ]
    behave_command[0] = command_path(venv_path, behave_command)
    behave_result = run_command(behave_command)

    if behave_result.returncode != 0:
        raise AcceptanceTestsError("behave tests run failed")



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import contextlib
import dataclasses
import io
import sys
import typing


@dataclasses.dataclass
class StdIo:
    err: io.StringIO
    out: io.StringIO


@contextlib.contextmanager
def capture_io() -> typing.ContextManager[StdIo]:
    new = StdIo(err=io.StringIO(), out=io.StringIO())
    with contextlib.redirect_stderr(new.err), contextlib.redirect_stdout(
        new.out
    ):
        yield new



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import subprocess

import pytest

import build_harness.commands.build_harness_group
from build_harness.commands import ExitState, main
from build_harness.commands.bdd.run_tests import AcceptanceTestsError, _run_bdd
from tests.ci.support.click_runner import click_runner  # noqa: F401

MOCK_VENV_PATH = pathlib.Path("/some/test_bdd/path")
EXPECTED_BEHAVE_PATH = str(MOCK_VENV_PATH / "behave")


@pytest.fixture()
def clean_exists_tests(mocker):
    mocker.patch.object(pathlib.Path, "exists", return_value=False)
    mocker.patch.object(pathlib.Path, "mkdir")


@pytest.fixture()
def clean_dir_tests(mocker):
    mocker.patch.object(pathlib.Path, "exists", return_value=True)
    mocker.patch.object(pathlib.Path, "is_dir", return_value=True)


@pytest.fixture()
def not_dir_tests(mocker):
    mocker.patch.object(pathlib.Path, "exists", return_value=True)
    mocker.patch.object(pathlib.Path, "is_dir", return_value=False)


class TestRunBdd:
    def test_disabled(self, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            return_value=mock_result,
        )

        _run_bdd(MOCK_VENV_PATH, False, pathlib.Path("some/xml/path"), list())

        mock_run.assert_called_once_with(
            [
                EXPECTED_BEHAVE_PATH,
                "--format",
                "pretty",
                "--tags",
                "~@notimplemented",
                "--tags",
                "~@nottestable",
                "features",
            ]
        )

    def test_enabled_nonexistent_dir(self, clean_exists_tests, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            return_value=mock_result,
        )

        _run_bdd(MOCK_VENV_PATH, True, pathlib.Path("some/xml/path"), list())

        mock_run.assert_called_once_with(
            [
                EXPECTED_BEHAVE_PATH,
                "--format",
                "pretty",
                "--junit",
                "--junit-directory",
                "some/xml/path",
                "--tags",
                "~@notimplemented",
                "--tags",
                "~@nottestable",
                "features",
            ]
        )

    def test_enabled_existent_dir(self, clean_dir_tests, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            return_value=mock_result,
        )

        _run_bdd(MOCK_VENV_PATH, True, pathlib.Path("dist/behave"), list())

        mock_run.assert_called_once_with(
            [
                EXPECTED_BEHAVE_PATH,
                "--format",
                "pretty",
                "--junit",
                "--junit-directory",
                "dist/behave",
                "--tags",
                "~@notimplemented",
                "--tags",
                "~@nottestable",
                "features",
            ]
        )

    def test_tag_list(self, clean_dir_tests, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            return_value=mock_result,
        )

        _run_bdd(
            MOCK_VENV_PATH,
            True,
            pathlib.Path("dist/behave"),
            ["@some.tag", "@other.tag,@this.thing"],
        )

        mock_run.assert_called_once_with(
            [
                EXPECTED_BEHAVE_PATH,
                "--format",
                "pretty",
                "--junit",
                "--junit-directory",
                "dist/behave",
                "--tags",
                "~@notimplemented",
                "--tags",
                "~@nottestable",
                "--tags",
                "@some.tag",
                "--tags",
                "@other.tag,@this.thing",
                "features",
            ]
        )

    def test_enabled_non_dir(self, not_dir_tests):
        with pytest.raises(
            AcceptanceTestsError,
            match=r"^JUnit XML location must be a " r"directory",
        ):
            _run_bdd(MOCK_VENV_PATH, True, pathlib.Path("dist/behave"), list())


class TestAcceptanceTests:
    def test_default(self, clean_exists_tests, click_runner, mocker):
        mock_behave_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_behave_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            side_effect=[mock_behave_result],
        )
        build_harness.commands.build_harness_group.sys.argv = [
            str(MOCK_VENV_PATH / "build-harness")
        ]
        result = click_runner.invoke(main, ["acceptance", "tests"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        EXPECTED_BEHAVE_PATH,
                        "--format",
                        "pretty",
                        "--tags",
                        "~@notimplemented",
                        "--tags",
                        "~@nottestable",
                        "features",
                    ]
                ),
            ]
        )

    def test_unknown_error(self, clean_exists_tests, click_runner, mocker):
        mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            side_effect=RuntimeError(),
        )
        result = click_runner.invoke(main, ["acceptance", "tests"])

        assert result.exit_code == ExitState.UNKNOWN_ERROR.value

    def test_junit_enabled(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        build_harness.commands.build_harness_group.sys.argv = [
            str(MOCK_VENV_PATH / "build-harness")
        ]
        result = click_runner.invoke(
            main, ["acceptance", "tests", "--junitxml"]
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        EXPECTED_BEHAVE_PATH,
                        "--format",
                        "pretty",
                        "--junit",
                        "--junit-directory",
                        "dist/behave",
                        "--tags",
                        "~@notimplemented",
                        "--tags",
                        "~@nottestable",
                        "features",
                    ]
                ),
            ]
        )

    def test_junit_alternate_dir(self, click_runner, mocker):
        expected_dir = "some/test_bdd/junit/dir"

        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        mocker.patch("build_harness.commands.bdd.run_tests._check_junit_dir")
        build_harness.commands.build_harness_group.sys.argv = [
            str(MOCK_VENV_PATH / "build-harness")
        ]
        result = click_runner.invoke(
            main,
            [
                "acceptance",
                "tests",
                "--junitxml",
                "--junitxml-dir",
                expected_dir,
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        EXPECTED_BEHAVE_PATH,
                        "--format",
                        "pretty",
                        "--junit",
                        "--junit-directory",
                        expected_dir,
                        "--tags",
                        "~@notimplemented",
                        "--tags",
                        "~@nottestable",
                        "features",
                    ]
                ),
            ]
        )

    def test_tag_list(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.run_tests.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        build_harness.commands.build_harness_group.sys.argv = [
            "/some/test_bdd/path/build-harness"
        ]
        result = click_runner.invoke(
            main,
            [
                "acceptance",
                "tests",
                "--tags",
                "@some.tag",
                "--tags",
                "@other.tag,@this.thing",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        EXPECTED_BEHAVE_PATH,
                        "--format",
                        "pretty",
                        "--tags",
                        "~@notimplemented",
                        "--tags",
                        "~@nottestable",
                        "--tags",
                        "@some.tag",
                        "--tags",
                        "@other.tag,@this.thing",
                        "features",
                    ]
                ),
            ]
        )


class TestSnippets:
    def test_clean(self, click_runner, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.snippets.run_command",
            side_effect=[mock_result],
        )
        build_harness.commands.build_harness_group.sys.argv = [
            str(MOCK_VENV_PATH / "build-harness")
        ]
        result = click_runner.invoke(main, ["acceptance", "snippets"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        EXPECTED_BEHAVE_PATH,
                        "--snippets",
                        "--tags",
                        "@notimplemented",
                        "features",
                    ]
                ),
            ]
        )


class TestTags:
    def test_clean(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.bdd.tags.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        build_harness.commands.build_harness_group.sys.argv = [
            str(MOCK_VENV_PATH / "build-harness")
        ]
        result = click_runner.invoke(main, ["acceptance", "tags"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        EXPECTED_BEHAVE_PATH,
                        "--format",
                        "tags.location",
                        "features",
                    ]
                ),
            ]
        )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import tempfile

import pytest

from build_harness import __version__
from build_harness.commands._publish_flow import (
    GitNotFoundError,
    PublishOptions,
    publish_flow_main,
)
from build_harness.commands.state import ExitState
from build_harness.tools.git import TagData
from tests.ci.support.click_runner import click_runner  # noqa: F401


class TestPublishFlowProject:
    def test_project(self, click_runner, mocker):
        expected_repo_path = "some/path"
        mock_get_tag = mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset=11),
        )

        result = click_runner.invoke(
            publish_flow_main, ["--project", expected_repo_path]
        )

        assert result.exit_code == 0
        mock_get_tag.assert_called_once_with(
            pathlib.Path(expected_repo_path), mocker.ANY
        )
        assert result.output == PublishOptions.test.name

    def test_bad_repo(self, click_runner):
        with tempfile.TemporaryDirectory() as this_dir:
            result = click_runner.invoke(
                publish_flow_main, ["--project", this_dir]
            )

            assert result.exit_code == ExitState.BAD_REPO.value
            assert "FAILED: Invalid git repository" in result.output

    def test_bad_git(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands._publish_flow.validate_git",
            side_effect=GitNotFoundError(
                "Git must be installed to use this utility"
            ),
        )
        result = click_runner.invoke(publish_flow_main, ["--git", "bad/path"])

        assert result.exit_code == ExitState.BAD_GIT_EXE.value
        assert (
            "Git must be installed and configured to use this utility"
            in result.output
        )
        assert (
            "sudo apt install -y git || sudo yum install git" in result.output
        )


class TestPublishFlowDefaultBranch:
    def test_default_branch(self, click_runner, mocker):
        expected_branch = "alternate_main"
        mock_get_tag = mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(tag="3.14.159"),
        )

        result = click_runner.invoke(
            publish_flow_main, ["--default-branch", expected_branch]
        )

        assert result.exit_code == 0
        mock_get_tag.assert_called_once_with(mocker.ANY, expected_branch)


class TestPublishFlowStockOptions:
    def test_help(self, click_runner):
        result = click_runner.invoke(publish_flow_main, ["--help"])

        assert result.exit_code == 0
        assert "Usage: publish-flow-main" in result.output

    def test_version(self, click_runner):
        result = click_runner.invoke(publish_flow_main, ["--version"])

        assert result.exit_code == 0
        assert __version__ in result.output


class TestPublishFlowActivation:
    def test_semver_release(self, click_runner, mocker):
        data = [
            {"arguments": [], "result": PublishOptions.yes.name},
            {
                "arguments": ["--publish-prerelease"],
                "result": PublishOptions.yes.name,
            },
            {
                "arguments": ["--publish-nonrelease"],
                "result": PublishOptions.yes.name,
            },
        ]
        mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset=None),
        )

        _do_publish_result_test(data, click_runner)

    def test_semver_prerelease(self, click_runner, mocker):
        data = [
            {"arguments": [], "result": PublishOptions.test.name},
            {
                "arguments": ["--publish-prerelease"],
                "result": PublishOptions.yes.name,
            },
            {
                "arguments": ["--publish-nonrelease"],
                "result": PublishOptions.yes.name,
            },
        ]
        mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(
                tag="3.14.159-alpha.3",
                offset=None,
            ),
        )

        _do_publish_result_test(data, click_runner)

    def test_dryrun_tag(self, click_runner, mocker):
        """Dryun should result in "protected" job execution, so should never publish."""
        # and in case it does, pypi.org rejects the PEP-440 local suffix for upload.
        data = [
            {"arguments": [], "result": PublishOptions.dryrun.name},
            {
                "arguments": ["--publish-prerelease"],
                "result": PublishOptions.dryrun.name,
            },
            {
                "arguments": ["--publish-nonrelease"],
                "result": PublishOptions.dryrun.name,
            },
        ]
        mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(
                tag="3.14.159+dryrun",
                offset=None,
            ),
        )

        _do_publish_result_test(data, click_runner)

    def test_nonrelease(self, click_runner, mocker):
        data = [
            {"arguments": [], "result": PublishOptions.test.name},
            {
                "arguments": ["--publish-prerelease"],
                "result": PublishOptions.test.name,
            },
            {
                "arguments": ["--publish-nonrelease"],
                "result": PublishOptions.yes.name,
            },
        ]
        mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset="1"),
        )

        _do_publish_result_test(data, click_runner)


def _do_publish_result_test(data: list, this_runner):
    for this_data in data:
        result = this_runner.invoke(publish_flow_main, this_data["arguments"])

        if result.exit_code != 0:
            pytest.fail(
                "Non-zero exit, {0}".format(str(this_data["arguments"]))
            )
        if result.output != this_data["result"]:
            pytest.fail(
                "Unexpected output, {0}, {1}, {2}".format(
                    str(this_data["arguments"]),
                    result.output,
                    this_data["result"],
                )
            )


class TestPublishFlowDisablePullRequestPublish:
    def test_nonrelease(self, click_runner, mocker):
        data = [
            {"arguments": [], "result": PublishOptions.test.name},
            {
                "arguments": ["--disable-pr-publish", "merge_request_event"],
                "result": PublishOptions.dryrun.name,
            },
            {
                "arguments": ["--disable-pr-publish", "anyothervalue"],
                "result": PublishOptions.test.name,
            },
        ]
        mocker.patch(
            "build_harness.commands._publish_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset="10"),
        )

        _do_publish_result_test(data, click_runner)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import tempfile

from build_harness import __version__
from build_harness.commands._release_flow import (
    GitNotFoundError,
    release_flow_main,
)
from build_harness.commands.state import ExitState
from build_harness.tools.git import TagData
from tests.ci.support.click_runner import click_runner  # noqa: F401


class TestReleaseFlowMain:
    def test_clean(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands._release_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset="11"),
        )

        result = click_runner.invoke(release_flow_main, [])

        assert result.exit_code == 0
        assert result.output == "3.14.159-post.11"

    def test_semver_compliant_post(self, click_runner, mocker):
        """semantic version compliant post form"""
        mocker.patch(
            "build_harness.commands._release_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset="11"),
        )

        result = click_runner.invoke(release_flow_main, [])

        assert result.exit_code == 0
        assert result.output == "3.14.159-post.11"

    def test_default(self, click_runner, mocker):
        mock_get_tag = mocker.patch(
            "build_harness.commands._release_flow.get_tag_data",
            return_value=TagData(tag="3.14.159"),
        )
        result = click_runner.invoke(release_flow_main, [])

        assert result.exit_code == 0
        mock_get_tag.assert_called_once_with(pathlib.Path("."), "master")
        assert result.output == "3.14.159"

    def test_existing_post_overwritten(self, click_runner, mocker):
        mock_get_tag = mocker.patch(
            "build_harness.commands._release_flow.get_tag_data",
            return_value=TagData(tag="3.14.159-post.5", offset="11"),
        )
        result = click_runner.invoke(release_flow_main, [])

        assert result.exit_code == 0
        mock_get_tag.assert_called_once_with(pathlib.Path("."), "master")
        assert result.output == "3.14.159-post.11"

    def test_project(self, click_runner, mocker):
        expected_repo_path = "some/path"
        mock_get_tag = mocker.patch(
            "build_harness.commands._release_flow.get_tag_data",
            return_value=TagData(tag="3.14.159", offset=11),
        )

        result = click_runner.invoke(
            release_flow_main, ["--project", expected_repo_path]
        )

        assert result.exit_code == 0
        mock_get_tag.assert_called_once_with(
            pathlib.Path(expected_repo_path), mocker.ANY
        )
        assert result.output == "3.14.159-post.11"

    def test_bad_repo(self, click_runner):
        with tempfile.TemporaryDirectory() as this_dir:
            result = click_runner.invoke(
                release_flow_main, ["--project", this_dir]
            )

            assert result.exit_code == ExitState.BAD_REPO.value
            assert "FAILED: Invalid git repository" in result.output

    def test_bad_git(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands._release_flow.validate_git",
            side_effect=GitNotFoundError(
                "Git must be installed to use this utility"
            ),
        )
        result = click_runner.invoke(release_flow_main, ["--git", "bad/path"])

        assert result.exit_code == ExitState.BAD_GIT_EXE.value
        assert (
            "Git must be installed and configured to use this utility"
            in result.output
        )
        assert (
            "sudo apt install -y git || sudo yum install git" in result.output
        )

    def test_default_branch(self, click_runner, mocker):
        expected_branch = "alternate_main"
        mock_get_tag = mocker.patch(
            "build_harness.commands._release_flow.get_tag_data",
            return_value=TagData(tag="3.14.159"),
        )

        result = click_runner.invoke(
            release_flow_main, ["--default-branch", expected_branch]
        )

        assert result.exit_code == 0
        mock_get_tag.assert_called_once_with(mocker.ANY, expected_branch)

    def test_help(self, click_runner):
        result = click_runner.invoke(release_flow_main, ["--help"])

        assert result.exit_code == 0
        assert "Usage: release-flow-main" in result.output

    def test_version(self, click_runner):
        result = click_runner.invoke(release_flow_main, ["--version"])

        assert result.exit_code == 0
        assert __version__ in result.output



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
import logging

from build_harness._utility import extract_coverage, run_command


class TestRunCommand:
    def test_default(self, caplog, mocker):
        with caplog.at_level(logging.DEBUG, logger="build_harness._utility"):
            mock_run = mocker.patch("build_harness._utility.subprocess.run")
            command = ["something", "--option"]

            result = run_command(command)

            mock_run.assert_called_once_with(command, text=True)
            assert result == mock_run.return_value
            # default neither command nor arguments suppressed in logs
            assert "command to run," in caplog.text
            assert "command arguments," in caplog.text

    def test_command_logging(self, caplog, mocker):
        with caplog.at_level(logging.DEBUG, logger="build_harness._utility"):
            mock_run = mocker.patch("build_harness._utility.subprocess.run")
            command = ["something", "--option"]

            result = run_command(command, suppress_command_logging=False)

            mock_run.assert_called_once_with(command, text=True)
            assert result == mock_run.return_value
            assert "command to run," in caplog.text
            assert "command arguments," in caplog.text

    def test_suppress_command_logging(self, caplog, mocker):
        with caplog.at_level(logging.DEBUG, logger="build_harness._utility"):
            mock_run = mocker.patch("build_harness._utility.subprocess.run")
            command = ["something", "--option"]

            result = run_command(command, suppress_command_logging=True)

            mock_run.assert_called_once_with(command, text=True)
            assert result == mock_run.return_value
            assert "command to run," not in caplog.text
            assert "command arguments," in caplog.text

    def test_argument_logging(self, caplog, mocker):
        with caplog.at_level(logging.DEBUG, logger="build_harness._utility"):
            mock_run = mocker.patch("build_harness._utility.subprocess.run")
            command = ["something", "--option"]

            result = run_command(command, suppress_argument_logging=False)

            mock_run.assert_called_once_with(command, text=True)
            assert result == mock_run.return_value
            assert "command to run," in caplog.text
            assert "command arguments," in caplog.text

    def test_suppress_argument_logging(self, caplog, mocker):
        with caplog.at_level(logging.DEBUG, logger="build_harness._utility"):
            mock_run = mocker.patch("build_harness._utility.subprocess.run")
            command = ["something", "--option"]

            result = run_command(command, suppress_argument_logging=True)

            mock_run.assert_called_once_with(command, text=True)
            assert result == mock_run.return_value
            assert "command to run," in caplog.text
            assert "command arguments," not in caplog.text


class TestExtractCoverage:
    def test_all_caps(self):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
TOTAL                                  170     44    91%
"""

        result = extract_coverage(mock_captured_report)

        assert result == 91

    def test_camel_case(self):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
Total                                  170     44    90%
"""

        result = extract_coverage(mock_captured_report)

        assert result == 90

    def test_lower_case(self):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
total                                  170     44    90%
"""

        result = extract_coverage(mock_captured_report)

        assert result == 90

    def test_zero(self):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
total                                  170     44    0%
"""

        result = extract_coverage(mock_captured_report)

        assert result == 0

    def test_100(self):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
total                                  170     44    100%
"""

        result = extract_coverage(mock_captured_report)

        assert result == 100



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Static analysis subcommand implementation."""

import logging
import pathlib
import sys

import click

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)
from build_harness.project import acquire_source_dir

from .code_style import FormattingError, _apply_formatting
from .state import CommandState, ExitState

log = logging.getLogger(__name__)


FLAKE8_FORMAT_CMD: CommandArgs = ["flake8", "."]
MYPY_FORMAT_CMD: CommandArgs = ["mypy", "--show-error-codes", "."]

VALID_ANALYSIS_OPTIONS = [
    "all",
    "flake8",
    "mypy",
    "pydocstyle",
]


class StaticAnalysisError(Exception):
    """Problem occurred during static analysis."""


class Flake8Error(StaticAnalysisError):
    """Problem occurred with flake8 analysis."""


class MypyError(StaticAnalysisError):
    """Problem occurred with mypy analysis."""


class PydocstyleError(StaticAnalysisError):
    """Problem occurred with pydocstyle anslysis."""


def _run_flake8(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run flake8 analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If flake8 analysis fails.
    """
    click.echo("flake8 analysis...")
    FLAKE8_FORMAT_CMD[0] = command_path(venv_path, FLAKE8_FORMAT_CMD)
    FLAKE8_FORMAT_CMD[-1] = str(source_path)
    result = run_command(FLAKE8_FORMAT_CMD)

    if result.returncode != 0:
        raise Flake8Error("flake8 analysis failed")
    click.echo(click.style("flake8 analysis completed", fg="green"))


def _run_mypy(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run mypy analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If mypy analysis fails.
    """
    click.echo("mypy analysis...")
    MYPY_FORMAT_CMD[0] = command_path(venv_path, MYPY_FORMAT_CMD)
    MYPY_FORMAT_CMD[-1] = str(source_path)
    result = run_command(MYPY_FORMAT_CMD)

    if result.returncode != 0:
        raise MypyError("mypy analysis failed")
    click.echo(click.style("mypy analysis completed", fg="green"))


def _run_pydocstyle(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run pydocstyle analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If pydocstyle analysis fails.
    """
    click.echo(
        click.style(
            "WARNING: pydocstyle analysis is deprecated and will be removed in "
            "the next major release. Run flake8 analysis instead with "
            "flake8-docstrings plugin",
            fg="yellow",
        )
    )
    _run_flake8(source_path, venv_path)


def _apply_analysis(
    analysis_type: str, project_path: pathlib.Path, venv_path: pathlib.Path
) -> None:
    """
    Apply the specified analysis.

    Args:
        project_path: Path to project root directory.
        venv_path: Path to Python virtual environment.
    """
    source_dir = acquire_source_dir(project_path)

    analysis_methods = {
        "flake8": _run_flake8,
        "mypy": _run_mypy,
        "pydocstyle": _run_pydocstyle,
    }
    if analysis_type != "all":
        analysis_methods[analysis_type](project_path / source_dir, venv_path)
    else:
        # exclude deprecated `pydocstyle` analysis from `all`
        for this_method in [
            y for x, y in analysis_methods.items() if x != "pydocstyle"
        ]:
            this_method(project_path / source_dir, venv_path)


@click.command()
@click.option(
    "--analysis",
    default="all",
    help="Analysis type to apply (default all)",
    type=click.Choice(VALID_ANALYSIS_OPTIONS, case_sensitive=True),
)
@click.pass_context
def static_analysis(ctx: click.Context, analysis: str) -> None:
    """
    Run suite of Python static analysis.

    The suite currently comprises "flake8", "pydocstyle" and "mypy". Formatting
    is run before the static analysis to eliminate formatting complaints from
    static analysis tools.
    """
    try:
        command_state: CommandState = ctx.obj

        _apply_formatting(command_state.venv_path)
        _apply_analysis(
            analysis, command_state.project_path, command_state.venv_path
        )
    except FormattingError as e:
        report_console_error(str(e))
        sys.exit(ExitState.FORMATTING_FAILED.value)
    except Flake8Error as e:
        report_console_error(str(e))
        sys.exit(ExitState.FLAKE8_FAILED.value)
    except MypyError as e:
        report_console_error(str(e))
        sys.exit(ExitState.MYPY_FAILED.value)
    except PydocstyleError as e:
        report_console_error(str(e))
        sys.exit(ExitState.PYDOCSTYLE_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""PEP-440 regular expression patterns."""

import re

from .pep426 import Pep426Patterns


class Pep440Patterns:
    """PEP-440 regular expression patterns."""

    VALID_SPECIFIERS = r"(==|~=|>=|<=|!=|>|<)"
    VALID_VERSION_CHARACTERS = r"[0-9ab.\-_!postdevrc]+"
    SPECIFIER_SET_PATTERN = r"({0}\s*{1})".format(
        VALID_SPECIFIERS, VALID_VERSION_CHARACTERS
    )

    SPECIFIER_PATTERN = re.compile(
        r"^\s*{0}(\s*(?P<specifier_sets>({1},)*(\s*{1})?))?".format(
            Pep426Patterns.UNNORMALIZED_PACKAGE_NAME, SPECIFIER_SET_PATTERN
        )
    )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import os
import pathlib
import tempfile

from behave import given, then, when

from build_harness._utility import run_command
from build_harness.tools.git import DEFAULT_DEFAULT_BRANCH_NAME
from tests.ci.support.repo import (
    FeatureDryrunTagOnHead,
    FeatureDryrunTags,
    TagOnDefaultHeadOnFeature,
    TagOnHead,
)

from .support import (
    _install_working_build_harness,
    chdir_project_dir,
    create_venv,
    git_repo_context,
)

DEBUG_INSTALL = False


@given("an environment without git installed")
def step_impl(context):
    # Just check that an error message is emitted for a bad git path. Deeper analysis
    # is done by unit tests.
    context.given_arguments = ["--git", "bad/path"]
    context.tag = "3.1.4"
    context.expected_version = context.tag
    context.create_repo = TagOnHead(context.tag)


@then("emits an install git error message to stderr")
def step_impl(context):
    assert "User specified git invalid" in context.run_result.stderr
    assert (
        "Git must be installed and configured to use this utility"
        in context.run_result.stdout
    )


@given("a working directory that is not a valid git repo")
def step_impl(context):
    context.given_arguments = list()


@when("release flow runs in a working directory that is not a valid git repo")
def step_impl(context):
    with tempfile.TemporaryDirectory() as this_dir:
        working_dir = pathlib.Path(this_dir)
        working_venv = working_dir / ".venv"
        working_venvbin = working_venv / "bin"

        create_venv(working_venv)
        _install_working_build_harness(working_venvbin)
        print(f"working dir, {this_dir}")
        print(f"cwd before chdir, {os.getcwd()}")
        with chdir_project_dir(working_dir):
            print(f"cwd after chdir, {os.getcwd()}")

            work_release_flow = working_venvbin / "release-flow"
            this_command = [str(work_release_flow)] + context.given_arguments
            result = run_command(
                this_command,
                capture_output=True,
                text=True,
                universal_newlines=True,
            )

            print("result.stdout: {0}".format(result.stdout))
            print("result.stderr: {0}".format(result.stderr))

            context.run_result = result
            context.mock_project = working_dir


@then("emits an invalid git repo error message to stderr")
def step_impl(context):
    assert "Invalid git repository" in context.run_result.stderr


@given("a PEP-440 noncompliant tag is latest tag")
def step_impl(context):
    context.given_arguments = list()
    context.tag = "non_pep440_tag"
    context.expected_version = context.tag
    context.create_repo = TagOnHead(context.tag)


@then("emits a bad tag error to stderr")
def step_impl(context):
    print(f"tag error stderr: {context.run_result.stderr}")
    assert "Tags must be PEP-440 compliant" in context.run_result.stderr


@given("a PEP-440 compliant tag on default branch")
def step_impl(context):
    context.given_arguments = list()
    context.tag = "1!2.3a4.dev6"
    context.expected_version = context.tag
    context.create_repo = TagOnHead(context.tag)


@when("release flow extracts the tag data")
def step_impl(context):
    with git_repo_context(
        context.create_repo,
        debug_install=DEBUG_INSTALL,
        default_branch=DEFAULT_DEFAULT_BRANCH_NAME,
    ) as (
        mock_repo,
        this_rf_command,
    ):
        print(f"repo dir, {mock_repo}")
        print(f"cwd before chdir, {os.getcwd()}")
        with chdir_project_dir(mock_repo):
            print(f"cwd after chdir, {os.getcwd()}")

            this_command = [str(this_rf_command)] + context.given_arguments
            result = run_command(
                this_command,
                capture_output=True,
                text=True,
                universal_newlines=True,
            )

            print("result.stdout: {0}".format(result.stdout))
            print("result.stderr: {0}".format(result.stderr))

            context.run_result = result
            context.mock_project = mock_repo


@then("the utility emits the constructed package version to stdout")
def step_impl(context):
    print(
        "package version (actual, expected): "
        f"{context.run_result.stdout}, {context.expected_version}"
    )
    assert context.run_result.stdout == context.expected_version


@given("the tag excludes a post identifier")
def step_impl(context):
    context.given_arguments = list()
    context.tag = "1!2.3a4.dev6"
    context.expected_version = context.tag


@given("HEAD commit on a feature branch")
def step_impl(context):
    context.expected_version = "1!2.3a4-post.2.dev6"
    context.create_repo = TagOnDefaultHeadOnFeature(context.tag)


@given("the tag includes a post identifier")
def step_impl(context):
    context.given_arguments = list()
    context.tag = "1!2.3a4.post11.dev6"


@given("a dryrun tag on default branch")
def step_impl(context):
    context.given_arguments = list()
    context.tag = "1!2.3a4.post11.dev6+dryrun"
    # There's a single tag on default branch that is dryrun so expect the offset from
    # root commit.
    context.expected_version = "0.0.0-post.2"
    context.create_repo = TagOnDefaultHeadOnFeature(context.tag)


@given("a PEP-440 compliant release tag on default branch")
def step_impl(context):
    context.given_arguments = list()
    context.tag = "1!2.3a4.dev6"


@given("dryrun tag in feature commit history on HEAD commit")
def step_impl(context):
    # Expect the dryrun tags to be overlooked to find the latest release tag.
    context.expected_version = "3.1+dryrun2"
    context.create_repo = FeatureDryrunTagOnHead(context.tag)


@given("dryrun tags in feature commit history prior to HEAD commit")
def step_impl(context):
    # Expect the dryrun tags to be overlooked to find the latest release tag.
    context.expected_version = "1!2.3a4-post.4.dev6"
    context.create_repo = FeatureDryrunTags(context.tag)



#  Copyright (c) {{cookiecutter.creation_year}} {{cookiecutter.author}}
#
#  This file is part of {{cookiecutter.project_name}}.

"""{{cookiecutter.project_summary}}"""

from ._version import __version__  # noqa: F401



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""CLI command processing for ``build-harness`` command."""

import pathlib
import sys

import click

from build_harness._logging import LoggingState
from build_harness._version import acquire_version

from ._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_DISABLED,
    DEFAULT_LOG_LEVEL,
    DEFAULT_PROJECT_PATH,
    VALID_LOG_LEVELS,
)
from .analysis import static_analysis
from .bdd import bdd_acceptance_command
from .bootstrap import bootstrap
from .code_style import formatting
from .dependencies import install
from .publishing import publish
from .state import CommandState
from .unit_tests import unit_test
from .wheel import package


@click.group()
@click.pass_context
@click.version_option(version=acquire_version())
@click.option(
    "--log-enable-console",
    "enable_console_log",
    default=DEFAULT_CONSOLE_LOGGING_ENABLED,
    help="Log to console.",
    is_flag=True,
)
@click.option(
    "--log-disable-file",
    "disable_file_log",
    default=DEFAULT_FILE_LOGGING_DISABLED,
    help="Disable file logging.",
    is_flag=True,
)
@click.option(
    "--log-level",
    "log_level",
    default=DEFAULT_LOG_LEVEL,
    help="Select logging level to apply to all enabled log sinks.",
    type=click.Choice(VALID_LOG_LEVELS, case_sensitive=False),
)
def main(
    ctx: click.Context,
    disable_file_log: bool,
    enable_console_log: bool,
    log_level: str,
) -> None:
    """Build harness group."""
    logging_state = LoggingState(
        disable_file_logging=disable_file_log,
        enable_console_logging=enable_console_log,
        log_level_text=log_level,
    )
    ctx.obj = CommandState(
        logging_state=logging_state,
        project_path=pathlib.Path(DEFAULT_PROJECT_PATH),
        venv_path=pathlib.Path(sys.argv[0]).parent.absolute(),
    )


main.add_command(bdd_acceptance_command, name="acceptance")
main.add_command(bootstrap, name="bootstrap")
main.add_command(formatting, name="formatting")
main.add_command(install, name="install")
main.add_command(package, name="package")
main.add_command(publish, name="publish")
main.add_command(static_analysis, name="static-analysis")
main.add_command(unit_test, name="unit-test")



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import logging
import subprocess
import typing

log = logging.getLogger(__name__)

CommandArgs = typing.List[str]


def run_command(
    command: CommandArgs, **kwargs: typing.Any
) -> subprocess.CompletedProcess:
    """
    Run a system command using ``subprocess.run``.

    Args:
        command: List of command and arguments.
        **kwargs: Optional arguments passed through to ``subprocess.run``.

    Returns:
        Subprocess results.
    """
    log.debug("command to run, {0}".format(str(command)))
    log.debug("command arguments, {0}".format(str(kwargs)))
    result = subprocess.run(command, **kwargs)

    return result


def run_command_sequence(sequence: typing.List[str]) -> None:
    """
    Run through a shell command sequence.

    Args:
        sequence: List of commands to be executed.
    """
    for this_command in sequence:
        response = run_command(this_command.split())


def init_git(default_branch_name: str) -> None:
    """
    Initialize git repo in the new project directory.

    Args:
        default_branch_name: Name of default git branch.
    """
    commands = [
        f"git init --initial-branch={default_branch_name}",
    ]
    run_command_sequence(commands)


def commit_initial_files() -> None:
    """Commit initial set of file from template to source control."""
    # Generated files typically need reformatting to comply with PEP-8 formatting.
    commands = [
        "{{ cookiecutter._venv_bin }}/build-harness install",
        "{{ cookiecutter._venv_bin }}/build-harness formatting",
        "{{ cookiecutter._venv_bin }}/pre-commit install --install-hooks",
        "git add .",
    ]
    run_command_sequence(commands)
    # .split() is too crude for the whitespace containing argument here, so just use
    # `run_command` directly.
    run_command(["git","commit", "-m", "Initial files from template"])


def main() -> None:
    default_branch_name = "{{ cookiecutter.default_branch }}"

    init_git(default_branch_name)
    commit_initial_files()


if __name__ == "__main__":
    main()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Manage acquisition of project metadata."""

import pathlib
import typing

import toml
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from ._base_exception import BuildHarnessError
from .tools import Pep440Patterns, pep503_normalize


class DependencyValidationError(BuildHarnessError):
    """Problem has occurred validation a dependency package."""


class PyprojecttomlError(BuildHarnessError):
    """Problem has occurred loading ``pyproject.toml`` file."""


DEPENDENCY_TYPES = typing.Literal["runtime", "dev", "doc", "test"]


class ProjectDependencies(typing.TypedDict):
    """Project package dependencies dictionary type hint."""

    runtime: typing.List[str]
    dev: typing.List[str]
    doc: typing.List[str]
    test: typing.List[str]


def _acquire_pyprojecttoml_data(
    project_root: pathlib.Path,
) -> typing.MutableMapping[str, typing.Any]:
    """
    Load data from Python project ``pyproject.toml`` file.

    Args:
        project_root: Path to project root directory.

    Returns:
        Loaded TOML data
    Raises:
        PyprojecttomlError: if file cannot be loaded.
    """
    pyproject_toml_path = project_root / "pyproject.toml"
    if pyproject_toml_path.is_file():
        with pyproject_toml_path.open(mode="r") as f:
            content = f.read()
            toml_data = toml.loads(content)
    else:
        raise PyprojecttomlError(
            "Missing pyproject.toml file, {0}".format(pyproject_toml_path)
        )

    return toml_data


def acquire_source_dir(project_dir: pathlib.Path) -> str:
    """
    Acquire a project root source package from a project directory.

    Args:
        project_dir: Project root directory.

    Returns:
        Name of project root source package.
    """
    toml_data = _acquire_pyprojecttoml_data(project_dir)

    return toml_data["tool"]["flit"]["metadata"]["module"]


def acquire_project_dependencies(
    project_dir: pathlib.Path,
) -> ProjectDependencies:
    """
    Acquire Python project package dependencies from project definition.

    Args:
        project_dir: Root directory of project repo.

    Returns:
        Project package dependencies.
    """
    toml_data = _acquire_pyprojecttoml_data(project_dir)

    result: ProjectDependencies = {
        "runtime": [],
        "dev": [],
        "doc": [],
        "test": [],
    }

    if "requires" in toml_data["tool"]["flit"]["metadata"]:
        result["runtime"] += toml_data["tool"]["flit"]["metadata"]["requires"]

    if "requires-extra" in toml_data["tool"]["flit"]["metadata"]:
        for key in ["dev", "doc", "test"]:
            if key in toml_data["tool"]["flit"]["metadata"]["requires-extra"]:
                # mypy doesn't like dynamic types on TypedDict type hints
                result[
                    key  # type:ignore
                ] = toml_data[
                    "tool"
                ][
                    "flit"
                ][
                    "metadata"
                ]["requires-extra"][key]

    return result


T = typing.TypeVar("T", bound="ProjectDependencyAnalysis")


class ProjectDependencyAnalysis:
    """
    Query project dependencies for valid releases.

    https://www.python.org/dev/peps/pep-0440/#version-specifiers
    """

    __dependency_list: typing.List[str]

    def __init__(
        self: T,
        dependencies: ProjectDependencies,
        dependency_type: DEPENDENCY_TYPES,
    ) -> None:
        """
        Project release queries.

        Args:
            dependencies: Dictionary of project dependencies.
            dependency_type: Type of dependency to validate.
        """
        self.dependency_type = dependency_type
        self.dependencies = dependencies

        if self.dependency_type == "all":
            self.__dependency_list = []
            for x in self.dependencies.values():
                self.__dependency_list += x
        else:
            self.__dependency_list = self.dependencies[self.dependency_type]

    @property
    def packages(self: T) -> typing.List[str]:
        """List of project packages for the specified dependency type."""
        value = []
        for this_package in self.__dependency_list:
            this_match = Pep440Patterns.SPECIFIER_PATTERN.search(this_package)
            if this_match is not None:
                value.append(this_match.group("package_name"))

        return value

    def valid_release(self: T, package_name: str, release_id: str) -> bool:
        """
        Indicate if the specified  release is valid for listed dependencies.

        Assumes the dependencies have been recovered from pyproject.toml or
        similar. Raises an exception if the package does not exist within the
        dependency type specified at construction.

        NOTE: arbitrary equality is not supported; appears to not be supported
        by the ``packaging`` library.

        https://www.python.org/dev/peps/pep-0440/#arbitrary-equality

        Args:
            package_name:
            release_id:

        Returns:
            True if the package release is valid. False otherwise.
        Raises:
            DependencyValidationError: If the package is not present in the
            dependency list.
        """
        version_spec = Version(release_id)
        normalized_package_name_spec = pep503_normalize(package_name)

        is_valid_release = False
        found_in_dependencies = False
        for this_package in self.__dependency_list:
            this_match = Pep440Patterns.SPECIFIER_PATTERN.search(this_package)
            if this_match is not None:
                dependency_name = pep503_normalize(
                    this_match.group("package_name")
                )

                if normalized_package_name_spec == dependency_name:
                    found_in_dependencies = True
                    # validate the release id
                    specifier_list_text = this_match.group("specifier_sets")

                    specifier_sets: typing.Optional[SpecifierSet]
                    if not specifier_list_text:
                        specifier_sets = SpecifierSet()
                    else:
                        specifiers = (
                            specifier_list_text.split(",")
                            if specifier_list_text
                            else []
                        )

                        specifier_sets = SpecifierSet(specifiers[0])
                        for this_text in specifiers[1::]:
                            specifier_sets &= SpecifierSet(this_text)

                    is_valid_release = version_spec in specifier_sets

        if not found_in_dependencies:
            raise DependencyValidationError(
                "Package not found in dependencies, {0}".format(package_name)
            )

        return is_valid_release



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Package generation subcommand implementation."""

import copy
import logging
import pathlib
import sys
import typing

import click
import git  # type: ignore

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)
from build_harness.project import acquire_source_dir

from ._release_id import validate_release_id
from .state import CommandState, ExitState

log = logging.getLogger(__name__)

DEFAULT_VERSION_FILE = "VERSION"

FLIT_BUILD_CMD: CommandArgs = ["flit", "build"]


class PackagingError(Exception):
    """Problem occurred during packaging."""


class ReleaseEmptyOption(click.Option):
    """Optional release id with empty value."""

    empty_value = True


T = typing.TypeVar("T", bound="ReleaseValueOption")


class ReleaseValueOption(click.Option):
    """Fix the help for the _set suffix."""

    def get_help_record(self: T, ctx: click.Context) -> typing.Tuple[str, str]:
        """Customize help text."""
        help = typing.cast(
            typing.Tuple[str, str],
            super(ReleaseValueOption, self).get_help_record(ctx),
        )
        return (help[0].replace("_set ", " "),) + help[1:]


def _apply_release_id(release_id: str, version_file_path: pathlib.Path) -> None:
    """
    Apply the release id to the project source package directory.

    Args:
        release_id: Release id to be applied to source directory.
        project_path: Path to project directory.
    """
    validate_release_id(release_id)

    log.info(
        "Saving version file, {0} ({1})".format(version_file_path, release_id)
    )
    with version_file_path.open(mode="w") as f:
        f.write(release_id)

    try:
        repo = git.Repo(str(version_file_path.parent / ".."))
        repo.index.add([str(version_file_path)])
        repo.index.commit(str(version_file_path))
    except git.InvalidGitRepositoryError as e:
        # log the warning and ignore the error
        log.warning(str(e))


def _build_flit_package(venv_path: pathlib.Path) -> None:
    """
    Build packages using flit.

    Args:
        venv_path: Path to Python virtual environment.

    Raises:
        PackagingError: If package build exits non-zero.
    """
    flit_cmd = copy.deepcopy(FLIT_BUILD_CMD)
    flit_cmd[0] = command_path(venv_path, flit_cmd)
    result = run_command(flit_cmd)

    if any([x != 0 for x in [result.returncode]]):
        raise PackagingError("flit failed during package build.")


@click.command()
@click.pass_context
@click.option(
    "--release-id",
    default=None,
    help="PEP-440 release id to apply to package. [default: do nothing]",
    type=str,
)
def package(ctx: click.Context, release_id: typing.Optional[str]) -> None:
    """Build wheel, sdist packages."""
    try:
        command_state: CommandState = ctx.obj

        source_dir = pathlib.Path(
            acquire_source_dir(command_state.project_path)
        )
        version_file_path = source_dir / DEFAULT_VERSION_FILE

        try:
            if release_id:
                _apply_release_id(release_id, version_file_path)

            _build_flit_package(command_state.venv_path)
        except PackagingError as e:
            message = str(e)
            log.error(message)
            report_console_error(message)
            sys.exit(ExitState.PACKAGING_FAILED.value)
        finally:
            version_file_path.unlink(missing_ok=True)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import contextlib
import pathlib
import tempfile
import typing

import git
import pytest

from build_harness.tools.git import (
    GitError,
    GitNotFoundError,
    GitRepoError,
    TagData,
    TagDataError,
    _is_commit_tagged,
    _parse_describe,
    _validate_git_version,
    get_tag_data,
    validate_git,
)
from tests.ci.support.repo import (
    FeatureDryrunTagOnHead,
    FeatureDryrunTags,
    NoTags,
    TagBase,
    TagOnDefaultHeadOnFeature,
    TagOnHead,
)


class TestValidateGitVersion:
    def test_clean(self, mocker):
        mock_result = mocker.Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        mock_result.stdout = "git version 2.32.1 (Apple Git-133)"

        mock_run = mocker.patch(
            "build_harness.tools.git.subprocess.run", return_value=mock_result
        )

        _validate_git_version()

    def test_minimum_pass(self, mocker):
        mock_result = mocker.Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        mock_result.stdout = "git version 2.28"

        mock_run = mocker.patch(
            "build_harness.tools.git.subprocess.run", return_value=mock_result
        )

        _validate_git_version()

    def test_minimum_fail(self, mocker):
        mock_result = mocker.Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        mock_result.stdout = "git version 2.27.4"

        mock_run = mocker.patch(
            "build_harness.tools.git.subprocess.run", return_value=mock_result
        )

        with pytest.raises(GitError, match=r"^Incorrect git version must be"):
            _validate_git_version()


class TestValidateGit:
    def test_system_clean(self, mocker):
        mocker.patch("build_harness.tools.git._validate_git_version")
        mocker.patch(
            "build_harness.tools.git.shutil.which",
            return_value="/usr/bin/git",
        )

        validate_git(None)

    def test_system_bad(self, mocker):
        mocker.patch("build_harness.tools.git._validate_git_version")
        mocker.patch("build_harness.tools.git.shutil.which", return_value=None)

        with pytest.raises(
            GitNotFoundError, match=r"^Git not available on users PATH"
        ):
            validate_git(None)

    def test_user_precedence(self, mocker):
        mocker.patch("build_harness.tools.git._validate_git_version")
        mocker.patch(
            "build_harness.tools.git.shutil.which",
            return_value="/usr/bin/git",
        )
        mock_exists = mocker.patch(
            "build_harness.tools.git.os.path.exists", return_value=True
        )

        validate_git("some/other/path")

        mock_exists.assert_called_once_with("some/other/path")

    def test_user_clean(self, mocker):
        mocker.patch("build_harness.tools.git._validate_git_version")
        mocker.patch("build_harness.tools.git.shutil.which", return_value=None)
        mocker.patch(
            "build_harness.tools.git.os.path.exists", return_value=True
        )

        validate_git("some/other/path")

    def test_user_bad(self, mocker):
        mocker.patch("build_harness.tools.git._validate_git_version")
        mocker.patch("build_harness.tools.git.shutil.which", return_value=None)

        with pytest.raises(
            GitNotFoundError, match=r"^User specified git invalid"
        ):
            validate_git("some/other/path")


class SimpleTags(TagBase):
    def __init__(self, tag: str):
        super().__init__(tag)

    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path):
        self._commit_file(repo_dir / "c1", repo)
        repo.create_tag(self.tag, message="default branch tag")
        self._commit_file(repo_dir / "c2", repo)


@contextlib.contextmanager
def temp_repo(
    repo_init: TagBase, initial_branch: str = "main"
) -> typing.Generator[git.Repo, None, None]:
    with tempfile.TemporaryDirectory() as dir:
        repo_path = pathlib.Path(dir)
        repo = git.Repo.init(str(repo_path), initial_branch=initial_branch)
        repo_init(repo, repo_path)

        yield repo, repo_path


class TestIsCommitTagged:
    def test_true(self):
        with temp_repo(SimpleTags("some_tag")) as (git_repo, repo_path):
            assert _is_commit_tagged(git_repo.commit("HEAD^"), git_repo)

    def test_false(self):
        with temp_repo(SimpleTags("some_tag")) as (git_repo, repo_path):
            assert not _is_commit_tagged(git_repo.head.commit, git_repo)


class TestParseDescribe:
    def test_semantic_tag(self):
        mock_describe = "3.1.4"

        result = _parse_describe(mock_describe)

        assert result == TagData(tag="3.1.4", offset=None)

    def test_semantic_tag_offset(self):
        mock_describe = "3.1.4-14-gc3ba625"

        result = _parse_describe(mock_describe)

        assert result == TagData(tag="3.1.4", offset="14")

    def test_pep440_tag_offset(self):
        mock_describe = "3!1.4.15.92a6.post5.dev4-14-gc3ba625"

        result = _parse_describe(mock_describe)

        assert result == TagData(tag="3!1.4.15.92a6.post5.dev4", offset="14")

    def test_no_match(self, mocker):
        mock_describe = "3!1.4.15.92a6.post5.dev4-some-stuff"

        # The regex pattern seems to be quite strong, so not an obvious path to an
        # error here. Just use a patch to force the error state.
        mocker.patch("build_harness.tools.git.re.search", return_value=None)

        with pytest.raises(TagDataError):
            _parse_describe(mock_describe)


class TestGetTagData:
    def test_semantic_tag(self):
        expected_tag_data = TagData(tag="1!2.3a4.dev6", offset=None)

        with temp_repo(TagOnHead("1!2.3a4.dev6")) as (git_repo, repo_path):
            result = get_tag_data(repo_path)

        assert result == expected_tag_data

    def test_dryrun_on_head(self):
        """Dryrun tag on HEAD gets used as-is."""
        expected_tag_data = TagData(tag="3.1+dryrun2")

        with temp_repo(FeatureDryrunTagOnHead("1!2.3a4.dev6")) as (
            git_repo,
            repo_path,
        ):
            result = get_tag_data(repo_path)

        assert result == expected_tag_data

    def test_dryrun_in_history(self):
        """Dryrun tags in commit history get ignored."""
        expected_tag_data = TagData(tag="1!2.3a4.dev6", offset="4")

        with temp_repo(FeatureDryrunTags("1!2.3a4.dev6")) as (
            git_repo,
            repo_path,
        ):
            result = get_tag_data(repo_path)

        assert result == expected_tag_data

    def test_nondefault_branch(self):
        expected_tag_data = TagData(tag="0.0.0", offset="1")

        with temp_repo(NoTags()) as (git_repo, repo_path):
            nb = git_repo.create_head("other_branch")
            git_repo.head.reference = nb
            git_repo.head.reset(index=True, working_tree=True)
            try:
                git_repo.delete_head(git_repo.heads.main)
            except AttributeError:
                # this exception is raised on an attempt to delete a head that
                # doesn't exist. this test is just trying to define the
                # appropriate entry state for the test, so just swallow this
                # error.
                pass

            result = get_tag_data(repo_path, default_branch_name="other_branch")

        assert result == expected_tag_data

    def test_semantic_tag_offset(self):
        expected_tag_data = TagData(tag="3.1.4", offset="2")
        with temp_repo(TagOnDefaultHeadOnFeature("3.1.4")) as (
            git_repo,
            repo_path,
        ):
            result = get_tag_data(repo_path)

        assert result == expected_tag_data

    def test_no_tag(self):
        default_branch = "main"
        expected_tag_data = TagData(tag="0.0.0", offset="1")

        with temp_repo(NoTags(), initial_branch=default_branch) as (
            git_repo,
            repo_path,
        ):
            result = get_tag_data(repo_path, default_branch_name=default_branch)

        assert result == expected_tag_data

    def test_bad_tag(self):
        with tempfile.TemporaryDirectory() as this_dir:
            mock_path = pathlib.Path(this_dir)

            with pytest.raises(GitRepoError, match=r"^Invalid git repository"):
                get_tag_data(mock_path)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Common constants for command definitions."""

import enum

DEFAULT_PROJECT_PATH = "."

DEFAULT_CONSOLE_LOGGING_ENABLED = False
DEFAULT_FILE_LOGGING_DISABLED = False
DEFAULT_FILE_ROTATION_BACKUPS = 10
DEFAULT_FILE_ROTATION_ENABLED = True
DEFAULT_FILE_ROTATION_SIZE_MB = 1
DEFAULT_LOG_FILE = "build_harness.log"

DEFAULT_LOG_LEVEL = "warning"
VALID_LOG_LEVELS = ["critical", "error", "warning", "info", "debug", "notset"]


@enum.unique
class PublishOptions(enum.Enum):
    """Enumeration of artifact publish options."""

    yes = enum.auto()
    no = enum.auto()
    dryrun = enum.auto()
    test = enum.auto()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""
CLI argument and command processing.

Exports:

*  ``ExitState`` class
*  ``main`` function
*  ``publish_flow_main`` function
*  ``release_flow_main`` function
"""

from ._publish_flow import publish_flow_main  # noqa: F401
from ._release_flow import release_flow_main  # noqa: F401
from .build_harness_group import main  # noqa: F401
from .state import ExitState  # noqa: F401



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Package version management."""

import pathlib

from ._default_values import DEFAULT_RELEASE_ID


def acquire_version() -> str:
    """
    Acquire PEP-440 compliant version from VERSION file.

    Returns:
        Acquired version text.
    Raises:
        RuntimeError: If version is not valid.
    """
    here = pathlib.Path(__file__).parent
    version_file_path = (here / "VERSION").absolute()

    if version_file_path.is_file():
        with version_file_path.open(mode="r") as version_file:
            version = version_file.read().strip()
    else:
        version = DEFAULT_RELEASE_ID

    if not version:
        raise RuntimeError("Unable to acquire version")

    return version


__version__ = acquire_version()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import contextlib
import os
import pathlib
import tempfile
import typing

import git

from build_harness._utility import run_command


@contextlib.contextmanager
def chdir_project_dir(
    path: pathlib.Path,
) -> typing.Generator[None, pathlib.Path, None]:
    current_dir = pathlib.Path(".").absolute()
    try:
        os.chdir(path)

        yield path
    finally:
        os.chdir(str(current_dir))


def create_venv(venv_path: pathlib.Path) -> pathlib.Path:
    result = run_command(
        ["python3", "-m", "venv", str(venv_path)],
        capture_output=True,
        universal_newlines=True,
    )
    assert result.returncode == 0

    pip_path = venv_path / "bin" / "pip"
    assert pip_path.is_file()

    return pip_path


def _install_working_build_harness(working_venvbin: pathlib.Path) -> None:
    result = run_command(
        [str(working_venvbin / "pip"), "install", "flit"],
        capture_output=True,
        universal_newlines=True,
    )
    print(result.stdout)
    print(result.stderr)

    assert result.returncode == 0

    result = run_command(
        [str(working_venvbin / "flit"), "install", "-s"],
        capture_output=True,
        universal_newlines=True,
    )
    print(result.stdout)
    print(result.stderr)

    assert result.returncode == 0

    assert (working_venvbin / "build-harness").is_file()


class MockTempDir:
    THIS_COUNT = 0
    name: str

    def __init__(self, debug_install: bool) -> None:
        self._debug_install = debug_install
        self._path = pathlib.Path(__file__).parent / "data/{0}".format(
            MockTempDir.THIS_COUNT
        )
        self._path.mkdir(exist_ok=True, parents=True)
        MockTempDir.THIS_COUNT += 1
        self.name = str(self._path)

    def cleanup(self) -> None:
        # Do nothing because this is for debug and in this case we want the
        # "temporary" workspace to persist for analysis.
        if not self._debug_install:
            self._path.rmdir()


@contextlib.contextmanager
def build_harness_context(
    create_pyproject: typing.Callable[[pathlib.Path], None],
    release_id: str,
    existing_project_venv: bool = False,
    debug_install: bool = False,
    pre_step_actions: typing.Optional[typing.Callable[[dict], None]] = None,
) -> typing.Generator[typing.Tuple[pathlib.Path, pathlib.Path], None, None]:
    temporary_dir = None
    try:
        if debug_install:
            temporary_dir = MockTempDir(debug_install)
        else:
            temporary_dir = tempfile.TemporaryDirectory()

        this_project = pathlib.Path(temporary_dir.name)
        working_venv = this_project / "working"
        working_venvbin = working_venv / "bin"

        if existing_project_venv:
            create_venv(this_project / ".venv")

        pyproject_path = this_project / "pyproject.toml"
        create_pyproject(pyproject_path)

        create_venv(working_venv)
        _install_working_build_harness(working_venvbin)

        work_build_harness = working_venvbin / "build-harness"

        if pre_step_actions:
            pre_step_actions(
                {"working_dir": this_project, "release_id": release_id}
            )

        yield (temporary_dir.name, work_build_harness)
    finally:
        if temporary_dir:
            temporary_dir.cleanup()


@contextlib.contextmanager
def git_repo_context(
    create_repo_commits: typing.Callable[[git.Repo], None],
    debug_install: bool = False,
    default_branch: str = "main",
) -> typing.Generator[typing.Tuple[pathlib.Path, pathlib.Path], None, None]:
    """
    Create a git repo context for a test.

    Args:
        create_repo_commits: Callable that creates a structure of commits in the repo
                             for a test.
        debug_install: Debug enabled creates the working directory locally and does
                       not erase it at the end of the test.
        default_branch: Default branch name for repo initialization.

    Yields:
        Tuple of paths to git repo directory and release-flow command respectively.
    """
    temporary_dir = None
    try:
        if debug_install:
            temporary_dir = MockTempDir(debug_install)
        else:
            temporary_dir = tempfile.TemporaryDirectory()

        working_dir = pathlib.Path(temporary_dir.name)
        working_venv = working_dir / ".venv"
        working_venvbin = working_venv / "bin"

        create_venv(working_venv)
        _install_working_build_harness(working_venvbin)

        r = git.Repo.init(str(working_dir), initial_branch=default_branch)
        create_repo_commits(r, working_dir)

        work_release_flow = working_venvbin / "release-flow"

        yield (working_dir, work_release_flow)
    finally:
        if temporary_dir:
            temporary_dir.cleanup()


def create_mock_pyproject(file_path: pathlib.Path) -> None:
    """Create a minimal flit based mock pyproject.toml file."""
    with file_path.open(mode="w") as f:
        f.write(
            """[build-system]
requires = ["flit_core >=2,<3"]
build-backend = "flit_core.buildapi"

[tool.flit.metadata]
module = "some_module"
author = "Some Name"
author-email = "name@somewhere.com"
"""
        )


def create_source_dir(data: dict) -> None:
    """Create a mock project source directory."""
    working_dir = data["working_dir"]
    release_id = data["release_id"]
    source_dir = working_dir / "some_module"
    source_dir.mkdir(parents=True)
    init_file = source_dir / "__init__.py"
    with init_file.open(mode="w") as f:
        f.write('"""Just a mock project"""\n')
        f.write(
            """
__version__ = "{0}"
""".format(
                release_id
            )
        )
